// MyList.cpp : implementation file
//
//
// SyncURLs Version 1.4
// Copyright  1998 Ziff-Davis Inc.
// First published in PC Magazine, US Edition
// January 6, 1998
// Written by Gregory A. Wolking
//

#include "stdafx.h"
#include "SyncURLs.h"
#include "MyList.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMyList

CMyList::CMyList()
{
}

CMyList::~CMyList()
{
}


BEGIN_MESSAGE_MAP(CMyList, CListCtrl)
	//{{AFX_MSG_MAP(CMyList)
	ON_NOTIFY_REFLECT(NM_CLICK, OnClick)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnDblclk)
	ON_NOTIFY_REFLECT(LVN_KEYDOWN, OnKeydown)
	ON_WM_CONTEXTMENU()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
	// Context menu command handler.
	// Note that command IDs are assigned consecutive values in resource.h
	// to allow using this command range mapping.
	ON_COMMAND_RANGE(ID_CONTEXT_SELECT_ALL, ID_CONTEXT_INVERT, OnContextCommand)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMyList message handlers

// Mouse click
void CMyList::OnClick(NMHDR* pNMHDR, LRESULT* pResult) 
{
	int		iItem;
	UINT	nFlags;
	CPoint	point;

	GetCursorPos(&point);
	ScreenToClient(&point);
	iItem = HitTest(point, &nFlags);
	if (iItem >= 0)
		if ((nFlags & LVHT_ONITEMSTATEICON) != 0)
			Toggle_Item(iItem);
	*pResult = 0;
}

// Called when:
//   User clicks an item's checkbox
//   User double-clicks an item's path or icon.
//   User presses spacebar when an item is selected. 
void CMyList::Toggle_Item(int iItem)
{
	LV_ITEM lvi;

	lvi.mask = LVIF_PARAM | LVIF_STATE;		// Set flags to retrieve item info.
	lvi.stateMask = LVIS_STATEIMAGEMASK;
	lvi.iItem = iItem;
	lvi.iSubItem = 0;
	GetItem(&lvi);							// Get the info.
	if (lvi.lParam & 0x80000000)			// If item is a header,
		return;								// nothing to do.
	if (lvi.state == 0x2000)	// Toggle checkbox state.
		lvi.state = 0x3000;
	else
		lvi.state = 0x2000;
	lvi.mask = LVIF_STATE;		// Set flag.
	SetItem(&lvi);				// Set new state.
	Update(iItem);				// Force control to redraw the item.
}

// Called when user double-clicks an item.
void CMyList::OnDblclk(NMHDR* pNMHDR, LRESULT* pResult) 
{
	int		iItem;
	CPoint	point;

	GetCursorPos(&point);		// Get mouse position.
	ScreenToClient(&point);		// Convert to client coordinates.
	iItem = HitTest(point);		// Get item (if any) the user clicked on.
	if (iItem >= 0)				// If user did click on an item,
		Toggle_Item(iItem);		// toggle its checkbox.
	*pResult = 0;
}

// Called when user presses a key.
void CMyList::OnKeydown(NMHDR* pNMHDR, LRESULT* pResult) 
{
	LV_KEYDOWN* pLVKeyDown = (LV_KEYDOWN*)pNMHDR;

	if (pLVKeyDown->wVKey == VK_SPACE)			// Was it spacebar?
	{
		int nCount = GetItemCount();			// Get # of items.
		for (int i = 0; i < nCount; i++)		// Walk the list.
		{
			if (GetItemState(i, LVIS_SELECTED) != 0)	// If item is selected,
			{
				Toggle_Item(i);							// Toggle its state.
				break;
			}
		}
	}
	*pResult = 0;
}

// Called when user either right-clicks us or presses either the
// "menu" key or Shift+F10 when we have keyboard focus.
void CMyList::OnContextMenu(CWnd* pWnd, CPoint point) 
{
	UINT	flags = TPM_LEFTALIGN;		// Default menu alignment.
	CRect	rectClient;
	CMenu	mnu;
	CPoint	ptMenu;


	// Default menu position is wherever the mouse is.
	GetCursorPos(&ptMenu);
	if (point.x == -1 && point.y == -1)			// Were we activated by the keyboard?
	{
		GetClientRect(&rectClient);				// Yes .. get client rect,
		ClientToScreen(&rectClient);			// convert to screen coords,
		rectClient.NormalizeRect();				// and normalize it.
		if (!rectClient.PtInRect(ptMenu))		// If mouse is not within our client rect,
		{
			ptMenu = rectClient.CenterPoint();	// put menu in the middle of our client area.
			flags = TPM_CENTERALIGN;
		}
	}
	flags |= TPM_LEFTBUTTON;	// Regardless of alignment, track the left mouse button.
	mnu.CreatePopupMenu();
	mnu.AppendMenu(MF_STRING, ID_CONTEXT_SELECT_ALL, "&Select All");
	mnu.AppendMenu(MF_STRING, ID_CONTEXT_SELECT_NONE, "&Deselect All");
	mnu.AppendMenu(MF_SEPARATOR);
	mnu.AppendMenu(MF_STRING, ID_CONTEXT_INVERT, "&Invert Selection");
	mnu.TrackPopupMenu(flags, ptMenu.x, ptMenu.y, this);
}

///////////////////////////////////////////
// Context Menu command handler.
// Note that the menu command IDs, defined in resource.h, must have sequential values
// to allow this mechanism to function.
void CMyList::OnContextCommand(UINT nID) 
{
	int			iItemCount;
	LV_ITEM		lvi;

	lvi.mask = LVIF_PARAM | LVIF_STATE;			// Set flags to retrieve item info.
	lvi.stateMask = LVIS_STATEIMAGEMASK;		// Only need state image bits.
	lvi.iSubItem = 0;							// Only interested in first "sub-item".
	iItemCount = GetItemCount();				// Get number of items in list.
	for (lvi.iItem = 0; lvi.iItem < iItemCount; lvi.iItem++)	// For each item in the list:
	{
		GetItem(&lvi);					// Retrieve info for current item.
		if (lvi.lParam & 0x80000000)	// Is it a header?
			continue;					// Yes .. skip it.
		switch(nID)
		{
		case ID_CONTEXT_SELECT_ALL:		// "Select All":
			if (lvi.state == 0x2000)			// If item is not selected,
			{
				lvi.state = 0x3000;				// Select it.
				SetItem(&lvi);					// Update it.
				Update(lvi.iItem);				// Redraw it.
			}
			break;
		case ID_CONTEXT_SELECT_NONE:	// "Deselect All"
			if (lvi.state == 0x3000)			// If item is selected,
			{
				lvi.state = 0x2000;				// Deselect it.
				SetItem(&lvi);					// Update it.
				Update(lvi.iItem);				// Redraw it.
			}
			break;
		case ID_CONTEXT_INVERT:			// "Invert Selection".
			if (lvi.state == 0x3000)			// If item is selected,
				lvi.state = 0x2000;				// deselect it.
			else								// Otherwise,
				lvi.state = 0x3000;				// Select it.
			SetItem(&lvi);						// Update it.
			Update(lvi.iItem);					// Redraw it.
			break;
		default:					// Should never get here!
			break;
		}
	}
}

// Ensures that both columns are visible when the control is resized.
void CMyList::OnSize(UINT nType, int cx, int cy) 
{
	int		iWidth0,
			iWidth1,
			iWidthClient;
	CRect	rectClient;

	CListCtrl::OnSize(nType, cx, cy);
	iWidth0 = GetColumnWidth(0);			// Get width of column 0.
	// If column width hasn't been set yet, GetColumnWidth() returns a
	// really huge number, so:
	if (iWidth0 < 2048)						// If first column with is valid:
	{
		iWidth1 = GetColumnWidth(1);		// Get width of second column.
		GetClientRect(&rectClient);			// Get client area.
		iWidthClient = rectClient.Width();	// Get width of client area.
		if (iWidth0 > iWidthClient)				// Is first column wider than the control?
		{
			iWidth0 = iWidthClient / 2;			// Yes -- size both columns equally.
			iWidth1 = iWidthClient - iWidth0;
			SetColumnWidth(0, iWidth0);
			SetColumnWidth(1, iWidth1);
		}
		else if ((iWidth0 + iWidth1) != iWidthClient)	// No .. do columns fill the control exactly?
		{
			iWidth1 = iWidthClient - iWidth0;			// No .. Adjust second column to fill the control.
			SetColumnWidth(1, iWidth1);
		}
	}
}
