// SUDoc.cpp : implementation of the CSUDoc class
//
//
// SyncURLs Version 1.4
// Copyright  1998 Ziff-Davis Inc.
// First published in PC Magazine, US Edition
// January 6, 1998
// Written by Gregory A. Wolking
//

#include "stdafx.h"
#include <intshcut.h>
#include <direct.h>
#include <errno.h>

#include "SyncURLs.h"
#include "StatDlg.h"
#include "MyList.h"
#include "WeedDlg.h"
#include "SUDoc.h"
#include "BrowsDlg.h"
#include "ReplaceDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// OLE interface pointer.
IMalloc* pMalloc;
// Buffer used for OLE string conversion.
WORD szWide[_MAX_PATH];
// Increment used for document array allocation/reallocation
#define DOC_ARRAY_INCREMENT 32

extern CSyncURLsApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CSUDoc

IMPLEMENT_DYNCREATE(CSUDoc, CDocument)

BEGIN_MESSAGE_MAP(CSUDoc, CDocument)
	//{{AFX_MSG_MAP(CSUDoc)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE, OnUpdateFileSave)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE_AS, OnUpdateFileSaveAs)
	ON_UPDATE_COMMAND_UI(ID_EXPORT_NS, OnUpdateExportNs)
	ON_COMMAND(ID_EXPORT_TEXT, OnExportText)
	ON_UPDATE_COMMAND_UI(ID_EXPORT_TEXT, OnUpdateExportText)
	ON_UPDATE_COMMAND_UI(ID_EXPORT_MSIE, OnUpdateExportMsie)
	ON_UPDATE_COMMAND_UI(ID_EXPORT_BOTH, OnUpdateExportBoth)
	ON_COMMAND(ID_FILE_MANUAL, OnFileManual)
	ON_UPDATE_COMMAND_UI(ID_EXPORT_MANUAL, OnUpdateExportManual)
	ON_UPDATE_COMMAND_UI(ID_EXPORT_NSC, OnUpdateExportNsc)
	ON_COMMAND(ID_VIEW_LOG, OnViewLog)
	ON_UPDATE_COMMAND_UI(ID_VIEW_LOG, OnUpdateViewLog)
	ON_COMMAND(ID_OPT_REPLACE, OnOptReplace)
	ON_COMMAND(ID_FILE_OPENSPECIAL, OnFileOpenspecial)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSUDoc construction/destruction

CSUDoc::CSUDoc()
{
	int			iPos;
	CString		strLine1,
				strLine2;

	m_nItemCount = 0;
	m_nDocArraySize = 0;
	m_bFirstInit = TRUE;
	m_pDocElements = NULL;
	m_nPathCount = 0;
	m_pPathList = NULL;
	m_pLogFile = NULL;
	m_pStatDlg = NULL;
	m_bNSPresent = FALSE;
	m_bIEPresent = FALSE;
	m_bNSCPresent = FALSE;
	m_bOpenSpecial = FALSE;
	m_strIEPath.Empty();
	m_strNSPath.Empty();
	m_strNSCUserName.Empty();
	m_strNSCPath.Empty();
	m_strNSPageTitle = "Bookmarks Exported by SyncURLs";
	m_strNSCPageTitle = "Bookmarks Exported by SyncURLs";
	// Get filename for import log.
	strLine1 = theApp.m_pszHelpFilePath;
	iPos = strLine1.ReverseFind('\\') + 1;
	strLine2 = strLine1.Left(iPos);
	strLine2 += "IMPLOG.TXT";
	m_strErrLog = strLine2;
	// Set Browser-related variables.
	Detect_Browsers();
	// Initialize IMalloc interface pointer.
	if (SHGetMalloc(&pMalloc) == E_FAIL)
	{
		pMalloc = NULL;
			AfxMessageBox("SHGetMalloc failed.", MB_ICONSTOP);
	}
	m_chReplace = theApp.GetProfileInt("Settings", "ReplaceChar", '$');
}

CSUDoc::~CSUDoc()
{
	if (pMalloc)
		pMalloc->Release();
}

/////////////////////////////////
// Standard Document function overrides

// Called at app startup, or when File/New command is issued.
BOOL CSUDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;
	// Cannot continue if constructor could not initialize OLE interface pointer.
	if (pMalloc == NULL)
		return FALSE;
	// If any browser was detected,
	if (m_bNSPresent || m_bIEPresent || m_bNSCPresent)
		// Must re-import to ensure that tree structure is in synch
		// with the Favorites folder.
		Do_Import();
	else	//Otherwise just report the error.
	{
		if (m_bFirstInit)
		{
			AfxMessageBox("Could not detect either MSIE or Netscape!", MB_ICONSTOP);
			m_bFirstInit = FALSE;
		}
		Add_Path("\\", TRUE, NULL);
	}
	return TRUE;		// Report success.
}

BOOL CSUDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	CString strMessage;

	m_bLoadFailed = FALSE;				// Clear error flags. These will be set appropriately
	m_bInternalLoadFail = FALSE;		// by CSUDoc::Serialize if the file couldn't be loaded.
	m_bGotSomeData = FALSE;
	if (!CDocument::OnOpenDocument(lpszPathName))
		return FALSE;
	if (m_bLoadFailed)
	{
		if (m_bInternalLoadFail)
			strMessage.Format("Internal error occurred while\n"
						"reading %s.\n\n", lpszPathName);
		else
			strMessage.Format("Error reading %s\n"
				"The file is either not a Bookmark Library,\n"
				"or it is damaged in some way.\n\n", lpszPathName);
		AfxMessageBox(strMessage, MB_ICONINFORMATION);
		if (m_bGotSomeData)
			AfxMessageBox(IDP_PARTIAL_LOAD_FAIL, MB_HELP | MB_ICONINFORMATION);
		else
			AfxMessageBox(IDP_TOTAL_LOAD_FAIL, MB_ICONINFORMATION);
	}
	if (!m_bOpenSpecial)
		Do_Import();	
	return (!m_bLoadFailed);
}

// Called as part of File/New and File/Open
void CSUDoc::DeleteContents() 
{
	UINT i;

	if (m_nItemCount)								// If element array is not empty,
	{
		// Serious problem if array pointer is already NULL!
		ASSERT(m_pDocElements != NULL);

		for (i=0; i < m_nItemCount; i++)			// For each element,
		{
			if (m_pDocElements[i].pPage)			// Delete page name string.
				delete m_pDocElements[i].pPage;
			if (m_pDocElements[i].pUrl)				// Delete URL string.
				delete m_pDocElements[i].pUrl;
			if (m_pDocElements[i].pComment)			// Delete comment string.
				delete m_pDocElements[i].pComment;
		}
		free (m_pDocElements);					// Delete the array.
		m_pDocElements = NULL;						// NULL the pointer.
		m_nItemCount = 0;							// Reset item count.
		m_nDocArraySize = 0;						// Reset document array size.
	}
	if (m_nPathCount)								// If path array is not empty,
	{
		// Serious problem if array pointer is already NULL!
		ASSERT(m_pPathList != NULL);

		for (i=0; i < m_nPathCount; i++)			// For each array element,
		{
			if (m_pPathList[i].pPath)				// Delete the path string.
				delete m_pPathList[i].pPath;
		}
		delete [] m_pPathList;						// Delete the path array.
		m_pPathList = NULL;							// NULL the array pointer.
		m_nPathCount = 0;							// Reset the path count.
	}
}


/////////////////////////////////////////////////////////////////////////////
// CSUDoc serialization

void CSUDoc::Serialize(CArchive& ar)
{
	UINT		i;
	CString		strBlank("");

	if (ar.IsStoring())
	{
		ar << m_nPathCount;						// Write number of paths.
		if (m_nPathCount > 0)					// Write path strings.
		{
			for (i = 0; i < m_nPathCount; i++)
				ar << *m_pPathList[i].pPath;
		}
		ar << m_nItemCount;						// Write # of bookmarks.
		if (m_nItemCount > 0)					// Write each bookmark:
		{
			for (i = 0; i < m_nItemCount; i++)
			{
				ar << *m_pDocElements[i].pPage;		// Write page name string.
				if (m_pDocElements[i].pUrl)			// If URL pointer is not NULL,
					ar << *m_pDocElements[i].pUrl;	// write the comment,
				else								// otherwise
					ar << strBlank;					// write empty string.
				if (m_pDocElements[i].pComment)		// Do the same for comment string.
					ar << *m_pDocElements[i].pComment;
				else
					ar << strBlank;
				ar << m_pDocElements[i].nPath;		// Write numeric fields.
				ar << m_pDocElements[i].nAdd;
				ar << m_pDocElements[i].nMod;
				ar << m_pDocElements[i].nVisit;
			}
		}
	}
	else
	{
		UINT		nPaths, nItems, nOffset;
		CString		strPage, strURL, strComment;
		LONG		nAdd, nMod, nVisit;
		UINT		*pnNewOffsets = NULL;

		try
		{
			ar >> nPaths;						// Get number of paths.
			pnNewOffsets = new UINT[nPaths];	// Create array of offsets.
			if (pnNewOffsets == NULL)			// Throw exception if creation failed.
				throw ((int) 3);
			for (i = 0; i < nPaths; i++)		// For each path:
			{
				ar >> strPage;					// Read the path string from the file.
				if (strPage.GetLength() > 1)	// Check for invalid path caused by bug in version 1.0
				{
					if (strPage[0] == '\\')
					{
						DeleteContents();
						m_bGotSomeData = FALSE;
						throw ((int) 4);
					}
				}
				if(!Add_Path(strPage, (i == 0), &pnNewOffsets[i]))	// Add to array and store its new offset.
					throw ((int) 1);			// Throw exception if it failed.
				else							// Otherwise,
					m_bGotSomeData = TRUE;		// Set flag.
			}
			ar >> nItems;						// Get number of items.
			for (i=0; i < nItems; i++)			// For each item,
			{
				ar >> strPage;			// Read page string,
				ar >> strURL;			// URL string,
				ar >> strComment;		// comment string,
				ar >> nOffset;			// original path offset,
				ar >> nAdd;				// add date,
				ar >> nMod;				// modified date,
				ar >> nVisit;			// and last visit date.
				if (!Add_Item(pnNewOffsets[nOffset], strPage, strURL,
					strComment.IsEmpty() ? NULL : (LPCTSTR) strComment,
					nAdd, nMod, nVisit, NULL))
					throw ((int) 1);	// Add to array using new path offset.
				else
					m_bGotSomeData = TRUE;
			}
		}
		catch (int e)		// Catch integer exception we might have thrown ourselvess.
		{
			m_bLoadFailed = TRUE;
			if (e != 4)
				m_bInternalLoadFail = TRUE;
		}
		catch (...)		// Catch any other exceptions.
		{
			m_bLoadFailed = TRUE;
		}
		if (pnNewOffsets != NULL)
			delete [] pnNewOffsets;		// Delete array of new offsets.
	}
}

/////////////////////////////////////////////////////////////////////////////
// CSUDoc diagnostics

#ifdef _DEBUG
void CSUDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CSUDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CSUDoc menu commands and UpdateUI handlers.

// "Export" menu, "Text only..." command.
void CSUDoc::OnExportText() 
{
	if (!Do_Export_Warning())
		return;
	Export_Text();
}

// "File" menu, "Manual Import" command.
void CSUDoc::OnFileManual() 
{
	BOOL		bFatal = FALSE;
	CString		strFileName;

	// Create commmon File Open dialog.
	CFileDialog dlg(TRUE, ".htm", "BOOKMARK.HTM", OFN_HIDEREADONLY, "HTML files (*.htm; *.html)|*.htm; *.html||", NULL);
	if (dlg.DoModal() != IDOK)	// Launch the dialog.
		return;					// Done if user didn't OK the dialog.
	if (!Start_Import_Log())	// Try to create and initialize the import log file.
	{
		AfxMessageBox(IDP_ERR_LOG_FAIL);	// Error message and quit if unable to initialize the log.
		return;
	}
	strFileName = dlg.GetPathName();	// Get user-specified filename.
	m_pStatDlg = new CStatDlg;			// Create and display the status dialog.
	m_pStatDlg->CenterWindow();
	m_pStatDlg->ShowWindow(SW_SHOW);
	m_pStatDlg->Show_Message("Importing Netscape bookmarks");
	m_pLogFile->WriteString("Beginning Netscape import\n\n");
	if (Parse_NS_File(strFileName) == IDP_IMPORT_FAIL)		// Process the file.
		bFatal = TRUE;							// Set flag if it failed.
	if (!m_bErrorsLogged)
		m_pLogFile->WriteString("Netscape import complete ... no errors.\n\n");
	Do_Import_Cleanup();			// Sort data, resolve duplicates, etc.
	SetModifiedFlag();				// Mark document as dirty.
	End_Import_Log();				// Close the log file.
	m_pStatDlg->DestroyWindow();	// Remove the status dialog.
	delete m_pStatDlg;
	m_pStatDlg = NULL;
	if (m_bErrorsLogged)			// Show "critical error" message if appropriate.
		Query_Show_Log(bFatal);
	UpdateAllViews(NULL);			// Refresh the display.
}

// "View" menu, "Import Log" command.
void CSUDoc::OnViewLog() 
{
	HINSTANCE	hInst;
	CString		strMessage;

	// Launch the default command associated with the .TXT extension.
	hInst = ShellExecute(NULL, NULL, m_strErrLog, NULL, NULL, SW_SHOWNORMAL);
	// Give message if launch failed.
	if ((UINT) hInst <= 32)
	{
		strMessage.Format("Failed to launch external application\n"
			"Either there is no program associated with the .TXT\n"
			"extension or the existing association is damaged.\n\n"
			"If you wish to view the error log manually,\n"
			"you'll find it at:\n%s", m_strErrLog);
		AfxMessageBox(strMessage, MB_ICONINFORMATION);
	}
}

// "Options" menu, "Replacement Character" command.
void CSUDoc::OnOptReplace() 
{
	CReplaceDlg dlg;							// Create the dialog.
	
	dlg.m_chReplace = m_chReplace;				// Give the dialog the current setting.
	if (dlg.DoModal() == IDOK)					// Launch the dialog. If user said "OK"...
	{
		if (dlg.m_chReplace != m_chReplace)		// and setting actually changed....
		{	
			m_chReplace = dlg.m_chReplace;		// Store the new setting and save it to the Registry.
			theApp.WriteProfileInt("Settings", "ReplaceChar", m_chReplace);
			// Let user know the consequences.
			AfxGetMainWnd()->MessageBox("For this change to take effect, you must either close\n"
				"and re-start SyncURLs, or use the File | New command\n"
				"to discard the current library and re-import your bookmarks.");
		}
	}
}

// "File" menu, "Open Special..." command.
void CSUDoc::OnFileOpenspecial() 
{
	// Alert user to consequences & give him a chance to bail.
	if (AfxGetMainWnd()->MessageBox(
				"Opening a library this way could cause you to\n"
				"lose some or all of your exising bookmarks.\n\n"
				"Are you sure you want to continue?",
				"Warning!",
				MB_YESNO | MB_ICONQUESTION)
			!= IDYES)
		return;
	m_bOpenSpecial = TRUE;		// Set flag to tell OnOpenDocument() not to import after loading.
	theApp.OnFileOpen();		// Call CWinApp::OnFileOpen() to prompt for filename, etc.
	m_bOpenSpecial = FALSE;		// Reset flag.
}

/////////////////////////////
// User Interface Update handlers.

void CSUDoc::OnUpdateViewLog(CCmdUI* pCmdUI) 
{
	CFile file;

	if (!file.Open(m_strErrLog, CFile::modeRead, NULL))
	{
		file.Close();
		pCmdUI->Enable(FALSE);
	}
	else
		pCmdUI->Enable(TRUE);
}

void CSUDoc::OnUpdateFileSave(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount > 0);	
}

void CSUDoc::OnUpdateFileSaveAs(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount > 0);
}

void CSUDoc::OnUpdateExportNs(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount > 0 && m_bNSPresent);
}

void CSUDoc::OnUpdateExportText(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount > 0);	
}

void CSUDoc::OnUpdateExportMsie(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount > 0 && m_bIEPresent);
}

void CSUDoc::OnUpdateExportBoth(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount > 0 && ((m_bNSPresent + m_bNSCPresent + m_bIEPresent) > 1));
}

void CSUDoc::OnUpdateExportManual(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nItemCount != 0);	
}

void CSUDoc::OnUpdateExportNsc(CCmdUI* pCmdUI) 
{
	if (m_bNSCPresent)
	{
		CString strText;
		strText.Format("Netscape &4.x (%s)", m_strNSCUserName);
		pCmdUI->SetText(strText);
	}
	pCmdUI->Enable(m_bNSCPresent && (m_nItemCount > 0));
}


/////////////////////////////////
// CSUDoc Helpers

// Adds a new bookmark to the document.
// Returns TRUE if either the bookmark was added succesfully
// or a duplicate record (same folder, page name, and url) is found.
// If requested, offset of the new or existing item is returned through the *pnOffset pointer.
BOOL CSUDoc::Add_Item(UINT nPathOffset,			// Offset of the path to this bookmark (required).
					  LPCTSTR lpszPage,			// Page name (required).
					  LPCTSTR lpszURL,			// URL (required).
					  LPCTSTR lpszComment,		// Comment (required, but may be NULL if no comment).
					  LONG nAdd,				// Date added (Netscape only, default = 0)
					  LONG nMod,				// Date modified (Netscape only, default = 0)
					  LONG nVisit,				// Date of last visit (Netscape only, default = 0)
					  UINT *pnOffset)			// Pointer to UINT to receive offset of new/existing item. Default = NULL.
{
	int			iLen1,
				iLen2;
	char		chTrail;
	CString		strTemp1,
				strTemp2;
	DOC_ELEMENT *pdeTemp = NULL;
	DOC_ELEMENT deNew = {0, NULL, NULL, NULL, 0, 0, 0};

	if (m_nItemCount)		// If list is not empty,
	{
		for (UINT i = 0; i < m_nItemCount; i++)	// See if matching entry exists
		{
			if (m_pDocElements[i].nPath != nPathOffset)						// In the same folder?
				continue;													// No ... keep looking.
			if (m_pDocElements[i].pPage->CompareNoCase(lpszPage) != 0)		// With same page name?
				continue;													// No ... keep looking.
			if (m_pDocElements[i].pUrl->Compare(lpszURL) == 0)				// Same URL?
			{																// Yes .. duplicate found!
				if (pnOffset)												// If requested,
					*pnOffset = i;											// return item's offset to caller.
				return TRUE;												// Indicate success.
			}
			else															// URL doesn't match, so
			{
																			// Check for case where URLs are identical except for a trailing slash character
				iLen1 = m_pDocElements[i].pUrl->GetLength();		// Length of existing URL
				iLen2 = strlen(lpszURL);							// Length of new URL
				if (abs(iLen1 - iLen2) == 1)						// If difference is one character,
				{
					if (iLen1 > iLen2)							// If existing URL is longer,
					{
						chTrail = m_pDocElements[i].pUrl->GetAt(iLen2);	// Get trailing character from existing
						strTemp1 = m_pDocElements[i].pUrl->Left(iLen2);	// USL, then strip it.
						strTemp2 = lpszURL;									// Copy new URL verbatim.
					}
					else										// Existing URL is shorter...
					{
						chTrail = lpszURL[iLen1];					// Get trailing char from new URL.
						strTemp1 = lpszURL;							// Strip it.
						strTemp2 = strTemp1.Left(iLen1);
						strTemp1 = *m_pDocElements[i].pUrl;		// Copy existing URL.
					}
					// If the extra character was a slash and the URLs match,
					if ((chTrail == '/') && (strTemp1 == strTemp2))
					{
						if (pnOffset)							// If requested,
							*pnOffset = i;						// return offset to caller.
						return TRUE;							// Indicate success.
					}
				}
			}
		}
	}
	deNew.nPath = nPathOffset;				// Save offset to path string.
	deNew.pPage = new CString(lpszPage);	// Create string for page name.
	if (deNew.pPage == NULL)				// If creation failed,
		return FALSE;						// Report error.
	deNew.pUrl = new CString(lpszURL);		// Create new CString for URL.
	if (deNew.pUrl == NULL)			// If creation failed,
	{
		delete deNew.pPage;			// delete string for page name
		return FALSE;				// and report error.
	}
	if (lpszComment == NULL)		// If comment pointer is NULL,
		deNew.pComment = NULL;		// just copy it.
	else							// Otherwise,
	{
		deNew.pComment = new CString(lpszComment);	// create new CString for comment.
		if (deNew.pComment == NULL)					// If creation failed,
		{
			if (deNew.pUrl)							// if URL string exists,
				delete deNew.pUrl;					// delete it.
			delete deNew.pPage;						// Delete page name string.
			return FALSE;							// Report failure.
		}
	}
	deNew.nAdd = nAdd;		// Copy numeric fields.
	deNew.nMod = nMod;
	deNew.nVisit = nVisit;

	if (m_nItemCount == m_nDocArraySize)	// Is document array full?
	{
		m_nDocArraySize += DOC_ARRAY_INCREMENT;			// Increase size to next increment.
		pdeTemp = (DOC_ELEMENT *) realloc(m_pDocElements, m_nDocArraySize * sizeof(DOC_ELEMENT));	// Create/reallocate the array.
		if (pdeTemp == NULL)				// If creation/expansion failed,
		{
			delete deNew.pPage;			// Delete page string.
			if (deNew.pUrl)
				delete deNew.pUrl;		// Delete URL string.
			if (deNew.pComment)
				delete deNew.pComment;
			return FALSE;				// Report failure to caller.
		}
		else							// Otherwise,
			m_pDocElements = pdeTemp;		// save new (possibly changed!) array pointer.
	}
	ASSERT(m_pDocElements != NULL);		// Serious problem if array pointer is NULL here!
	// Add new element to array.
	m_pDocElements[m_nItemCount] = deNew;

	if (pnOffset)						// If requested,
		*pnOffset = m_nItemCount;	// return index of new element to caller.
	++m_nItemCount;				// Update item count.
	return (m_bItemsAdded = TRUE);	// Report success.
}

// Adds a path to the path list, or returns offset of
// existing path string if specified path is already there.
//
// Parameters:
//	lpszPath	= pointer to ASCIIZ string containing the new path.
//  bLocked		= whether or not folder is to be "locked" in the tree.
//	pnOffset	= pointer to UINT for returned offset.
//
// Return Value: TRUE if successful, otherwise FALSE.
BOOL CSUDoc::Add_Path(LPCTSTR lpszPath, BOOL bLocked, UINT * pnOffset)
{
	UINT			i, nThisPath;
	PATH_ELEMENT	*ppeTemp = NULL;
	CString			*pstrNew = NULL;

	if (lpszPath == NULL)							// Fail if path pointer is NULL.
		return FALSE;
	try												// Create new CString containing path.
	{
		pstrNew = new CString(lpszPath);
	}
	catch (CMemoryException *e)		// Catch possible exception thrown by CString's constructor
	{
		e->Delete();		// Delete the exception object.
		pstrNew = NULL;		// NULL the pointer.
	}
	if (pstrNew == NULL)							// If creation failed,
		return FALSE;								// report failure to caller.
	if (m_nPathCount)								// If path array is not empty,
	{
		for (i = 0; i < m_nPathCount; i++)			// See if specified path is already in the array.
		{
			if (m_pPathList[i].pPath->CompareNoCase(lpszPath) == 0)	// If it is,
			{
				delete pstrNew;										// delete the new CString
				if (bLocked)						// If new path is locked,
				{
					if (!m_pPathList[i].bLocked)	// and existing path isn't,
					{
						m_bItemsChanged = TRUE;			// Set modified flag and
						m_pPathList[i].bLocked = TRUE;	// mark the existing path as locked.
					}
				}
				if (pnOffset)		// If requested,
					*pnOffset = i;	// give caller the offset to the existing string.
				return TRUE;		// Report success.
			}
		}
		// If path does not already exist, try to expand the array by one element.
		ppeTemp = (PATH_ELEMENT *) realloc(m_pPathList, (m_nPathCount + 1) * sizeof(PATH_ELEMENT));
	}
	else
		// Otherwise, create a new array.
		ppeTemp = (PATH_ELEMENT *) malloc(sizeof(PATH_ELEMENT));
	if (ppeTemp == NULL)			// If unable to create or expand the array,
	{
		delete pstrNew;			// delete the new string
		return FALSE;			// and report failure.
	}
	m_pPathList = ppeTemp;							// Save new (possibly changed!) array pointer.
	nThisPath = m_nPathCount;						// Save offset to the new record.
	++m_nPathCount;									// Update array count.
	m_pPathList[nThisPath].pPath = pstrNew;		// Store pointer to new string.
	m_pPathList[nThisPath].bProcessed = FALSE;	// Set remaining fields.
	m_pPathList[nThisPath].bLocked = bLocked;
	m_pPathList[nThisPath].nSelected = 2;
	if (pstrNew->Find('\\') == pstrNew->GetLength()-1)		// If path is only one level deep,
		m_pPathList[nThisPath].nParent = 0;					// its parent is the root.
	else													// Otherwise,
	{
		CString strScratch1, strScratch2;
		int		iPos;
		UINT	nNewPath;
		BOOL	bFound = FALSE;

		strScratch1 = pstrNew->Left(pstrNew->GetLength() - 1);	// Extract path to its parent.
		iPos = strScratch1.ReverseFind('\\') + 1;
		strScratch2 = pstrNew->Left(iPos);
		for (i = 1; i < m_nPathCount; i++)					// Find it in the path array.
		{
			if (strScratch2.CompareNoCase(*m_pPathList[i].pPath) == 0)
			{
				m_pPathList[nThisPath].nParent = i;		// When found, set offset.
				bFound = TRUE;
				break;
			}
		}
		if (!bFound)	// If the parent path was not found,
		{
			if (Add_Path(strScratch2, bLocked, &nNewPath))		// Try to add it.
				m_pPathList[nThisPath].nParent = nNewPath;		// If successful, set offset.
			else
				return FALSE;									// Otherwise, tell caller we died.
		}
	}
	if (pnOffset)							// If requested,
		*pnOffset = nThisPath;				// Give offset of new element to caller.
	m_bItemsAdded = (m_nPathCount > 1);
	return TRUE;								// Report success to caller.
}

// Imports MSIE bookmarks.
// Called recursively.
// On first call, "path" parm must be NULL indicating that
// we're processing the root "Favorites" folder.
// On subsequent calls, "path" parm is the relative path
// from the "Favorites" folder to the directory to be scanned.
// Returns FALSE on fatal error, otherwise TRUE.
BOOL CSUDoc::Scan_Folder(LPCTSTR lpszPath)
{
	BOOL					bResult,
							bForeignFilesFound;
	UINT					nOffset,
							nParent;
	LPSTR					pszURL = NULL;
	HANDLE					hFind;
	CString					strCurPath,
							strFullPath,
							strSearchSpec,
							strFullSpec,
							strPageName,
							strTemp;
	HRESULT					hResult;
	WIN32_FIND_DATA			fd;
	IPersistFile			*pIPF = NULL;
	IUniformResourceLocator	*pIURL = NULL;

	if (lpszPath)		// If not processing the root folder,
	{
		strCurPath = lpszPath;		// Set current path.
		strCurPath += '\\';			// Append trailing backslash.
		if (!Add_Path(strCurPath, FALSE, &nOffset))	// Add path to document.
		{
			Report_Fatal_Error("CSUDoc::Scan_Folder:\nFailed to create internal path object");
			return FALSE;
		}
		strFullPath = m_strIEPath + strCurPath;	//Get fully qualified path to the folder.
	}
	else			// Otherwise,
	{
		strCurPath.Empty();			// Current path is empty,
		nOffset = 0;				// it's the root folder
		strFullPath = m_strIEPath;	// and full path is the path to the "Favorites" folder itself.
	}
	strSearchSpec = strFullPath + "*.*";			// Search for all files and folders.
	fd.dwFileAttributes = FILE_ATTRIBUTE_NORMAL;	// Any file attributes.
	hFind = FindFirstFile(strSearchSpec, &fd);		// Begin the search.
	if (hFind != INVALID_HANDLE_VALUE)
	{
		bForeignFilesFound = FALSE;		// Reset flag.
		while(TRUE)
		{
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)		// Is current object a folder?
			{
				if ((strcmp(fd.cFileName, ".") != 0) &&				// Is it not "." or ".."?
					(strcmp(fd.cFileName, "..") != 0))
				{
					CString new_path = strCurPath + fd.cFileName;	// Build path to new folder
					if (!Scan_Folder(new_path))						// and process it via recursive call.
						return FALSE;								// Stop recursion if call failed!
				}
			}
			else
			{
				strPageName = fd.cFileName;							// Extract filename.
				strTemp = strPageName.Right(4);						// Check extension.
				if (strTemp.CompareNoCase(".URL") == 0)			// If it's a bookmark,
				{
					if (fd.dwFileAttributes & FILE_ATTRIBUTE_READONLY) // Is it read-only?
						bForeignFilesFound = TRUE;						// Yes .. set flag to lock this folder.
					// Instantiate an IUniformResourceLocator object.
					hResult = CoCreateInstance(CLSID_InternetShortcut,
						NULL, CLSCTX_INPROC_SERVER, IID_IUniformResourceLocator, (void **) &pIURL);
					if (SUCCEEDED(hResult))
					{
						// Get pointer to the object's IPersistFile interface.
						hResult = pIURL->QueryInterface(IID_IPersistFile, (void **) &pIPF);
						if (!SUCCEEDED(hResult))
							pIURL->Release();
					}
					if (pIPF == NULL || pIURL == NULL)
					{
						Report_Fatal_Error("CSUDoc::Scan_Folder:\nCould not obtain OLE interface pointer.");
						m_bErrorsLogged = TRUE;
						return FALSE;
					}
					strFullSpec = strFullPath + strPageName;		// Get fully qualified path.
					MultiByteToWideChar(CP_ACP, 0, strFullSpec, -1, szWide, _MAX_PATH);	// Convert to OLE string.
					if (pIPF->Load(szWide, STGM_READ) == S_OK)		// Try to load the .URL object.
					{
						if (pIURL->GetURL(&pszURL) == S_OK)			// If successful, retrieve the URL
						{
							if ((strPageName.GetLength()) - 4 <= 50)	// Truncate the bookmark name at 50 characters.
								strTemp = strPageName.Left(strPageName.GetLength() - 4);
							else
								strTemp = strPageName.Left(50);
							strTemp.TrimRight();
							bResult = Add_Item(nOffset, strTemp, pszURL, NULL);	// Add bookmark to the document.
							if (!bResult)
								Report_Fatal_Error("CSUDoc::ScanFolder:\nCould not create internal bookmark item.");
							pMalloc->Free(pszURL);
						}
						else
						{
							strTemp.Format("Could not extract URL from %s\n", strFullSpec);
							m_pLogFile->WriteString(strTemp);
							m_bErrorsLogged = TRUE;
						}
						pIPF->SaveCompleted(szWide);	// Make sure the file is released.
						if (!bResult)					// Bail out of we failed to add the item to the document.
						{
							pIPF->Release();
							pIURL->Release();
							return FALSE;
						}
					}
					else	// Report error if unable to load the file.
					{
						strTemp.Format("Failed to access file %s\n", strFullSpec);
						m_pLogFile->WriteString(strTemp);
						m_bErrorsLogged = TRUE;
					}
					pIPF->Release();	// Release and NULL the OLE interface pointers.
					pIURL->Release();
					pIPF = NULL;
					pIURL = NULL;
				}
				else   // File is not an Internet Shortcut....
				{
					bForeignFilesFound = TRUE;		// Set flag.
				}
			}
			if (!FindNextFile(hFind, &fd))	// Continue with next file.
				break;
		}
		if (bForeignFilesFound)						// Any non-.URL files found?
		{
			m_pPathList[nOffset].bLocked = TRUE;	// Yes .. mark folder as locked.
			if (nOffset != 0)								// If it's not the root folder,
			{
				nParent = m_pPathList[nOffset].nParent;		// Get its parent folder.
				while (nParent != 0)						// As long as it's not the root,
				{
					m_pPathList[nParent].bLocked = TRUE;		// Mark parent folder as locked.
					nParent = m_pPathList[nParent].nParent;		// Get next parent folder.
				}
			}
		}
		return TRUE;
	}
	Report_Fatal_Error("CSUDoc::Scan_Folder:\nFindFirstFile API failed.");
	return FALSE;
}


// Function determines whether or not the specified path
// has any unprocessed children. If so, returns TRUE
// and sets the UINT pointed to by the child parameter
// to the offset of the first child found.
// Note that functions that call this one must call
// Reset_Path_Flags() before processing begins.
BOOL CSUDoc::Has_Children(UINT nOffset, UINT * pnChild)
{
	ASSERT(pnChild != NULL);

	for (UINT i = 1; i < m_nPathCount; i++)			// For each folder accept the root,
	{
		if (!m_pPathList[i].bProcessed)				// If path has not been processed,
		{
			if (m_pPathList[i].nParent == nOffset)	// and its parent is the target folder,
			{
				*pnChild = i;						// return its offset to caller.
				return TRUE;						// Signal success.
			}
		}
	}
	return FALSE;		// Signal no unprocessed children found.
}

// Dumps data in text form. This is mainly for debugging purposes,
// although some folks may find this format useful.
void CSUDoc::Export_Text()
{
	UINT		i,
				j;
	BOOL		bFound;
	CString		strLine,
				strFileName;
	CStdioFile	file;

	CFileDialog dlg(FALSE, ".txt", "", OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"Text Files (*.txt)|*.txt||", NULL);
	dlg.m_ofn.lpstrTitle = "Export Text File";
	if (dlg.DoModal() != IDOK)
		return;
	strFileName = dlg.GetPathName();
	if (!file.Open(strFileName, CFile::modeCreate | CFile::modeWrite, NULL))
	{
		strLine.Format("Could not create or replace\n%s", strFileName);
		AfxMessageBox(strLine, MB_OK);
		return;
	}
	for (i=0; i < m_nPathCount; i++)
	{
		if (m_pPathList[i].nSelected == 1)
			continue;
		if (i==0)
			strLine = "Folder: Root";
		else
			strLine.Format("Folder: %s\n", *m_pPathList[i].pPath);
		file.WriteString(strLine);
		bFound = FALSE;
		for (j = 0; j < m_nItemCount; j++)
		{
			if (m_pDocElements[j].nPath == i)
			{
				bFound = TRUE;
				strLine.Format("  Title: %s\n", *m_pDocElements[j].pPage);
				file.WriteString(strLine);
				strLine.Format("    URL: %s\n", *m_pDocElements[j].pUrl);
				file.WriteString(strLine);
			}
		}
		if (!bFound)
			file.WriteString("  Folder contains no bookmarks.\n");
	}
	file.Close();
}


// Extract bookmarks from Netscape's BOOKMARK.HTM file.
UINT CSUDoc::Parse_NS_File(LPCTSTR lpszFileName)
{
	int			iPos;
	BOOL		bGotTitle = FALSE;
	UINT		nResult = 0;
	CString		strLine1,
				strLine2;
	CStdioFile	file;

	// First, make sure we can get access to the file.
	strLine1.Format("Opening %s\n", lpszFileName);
	m_pLogFile->WriteString(strLine1);
	if (!file.Open(lpszFileName, CFile::modeRead | CFile::typeText | CFile::shareDenyWrite, NULL))
	{
		strLine1.Format("CSUDoc::Parse_NS_File:\nCould not access %s", lpszFileName);
		Report_Fatal_Error(strLine1);
		return IDP_IMPORT_FAIL;
	}
	m_pFile = &file;	// Store pointer to the CFile object.
	// Extract the user's root folder title.
	while (TRUE)
	{
		if (!file.ReadString(strLine1))		// Read a line.
		{
			if (bGotTitle)					// No error if we retrieved the folder title.
				break;
			Report_Fatal_Error("CSUDoc::Parse_NS_File:\nUnexpected end of file");	// Report error if end of file.
			nResult = IDP_IMPORT_FAIL;
			goto ErrorExit;
		}
		if ((iPos = strLine1.Find("<H1>")) >= 0)	// Look for <H1> tag that marks the first folder title.
		{
			iPos += 4;
			strLine2 = strLine1.Mid(iPos);
			if ((iPos = strLine2.Find("</H1>")) == -1)
			{
				strLine2.Format("CSUDoc::Parse_NS_File:\nHTML error in:\n%s\nMissing </H1> tag. Processing aborted", strLine1);
				Report_Fatal_Error(strLine2);
				nResult = IDP_IMPORT_FAIL;
				goto ErrorExit;
			}
			m_strNSPageTitle = strLine2.Left(iPos);	// Extract the page title.
			bGotTitle = FALSE;
			break;
		}
	}
	// Now recurse our way through the file, reporting error if anything barfs.
	if ((nResult = Parse_NS_Folder()))
		goto ErrorExit;
	file.Close();
	m_pFile = NULL;
	return 0;
ErrorExit:
	file.Close();
	m_pFile = NULL;
	return nResult;
}

// This recursive routine parses the folders in Netscape's BOOKMARK.HTM file.
// The first time it is called, both parameters are NULL, indicating that
// we're starting with the root folder.
BOOL CSUDoc::Parse_NS_Folder(LPCTSTR pszOldPath, LPCTSTR pszNewPath)
{
	int			iPos1;
	BOOL		bInComment = FALSE,
				bGotURL = FALSE;
	UINT		nPathOffset,
				nResult;
	LONG		lAddDate, 
				lModDate,
				lVisitDate;
	LPSTR		pszNewURL = NULL;
	CString		strCurPath,
				strTemp1,
				strTemp2,
				strDebugMsg,
				strPage,
				strURL,
				strComment,
				strErrMsg;
	static UINT nBlankFolderCount = 0;


	if (pszOldPath == NULL)	// Root folder?
	{
		strCurPath.Empty();		// Path is blank.
	}
	else	//Otherwise,
	{
		// Check path length.
		if (strlen(pszOldPath) + strlen(pszNewPath) + m_iBasePathLen > 200)
		{
			// If it's too long, try to create our default "Relocated" path.
			if (!Add_Path("Relocated\\", FALSE))
			{
				Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nCould not create \"Relocated\" folder");
				return IDP_IMPORT_FAIL;
			}
			// If that worked, reset the parent of the current folder
			// to our "Relocated" folder, then make sure we can create a path for the folder itself.
			strCurPath = "Relocated\\";
			strCurPath += pszNewPath;
			if (!Add_Path(strCurPath, FALSE))
			{
				strErrMsg.Format("CSUDoc::Parse_NS_Folder:\nCould not create %s folder", strCurPath);
				Report_Fatal_Error(strErrMsg);
				return IDP_IMPORT_FAIL;
			}
			// Let the user know that we had to move the folder.
			strErrMsg.Format("Folder \"%s\"\n"
				"could not be created because the path is too long\n"
				"Folder was moved to a root folder named \"Relocated\"", pszNewPath);
			m_pLogFile->WriteString(strErrMsg);
			m_bErrorsLogged = TRUE;
		}
		else	// Path is not too long,
		{
			// So just tack the new folder name onto the existing one.
			strCurPath = pszOldPath;
			strCurPath += pszNewPath;
		}
	}
	// Search for <DL><p> tag, indicating the start of a folder's contents.
	while (TRUE)
	{
		if (!m_pFile->ReadString(strTemp1))	// Report failure if EOF reached.
		{
			Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nUnexpected end of file");
			return IDP_IMPORT_FAIL;
		}
		if (strTemp1.Find("<DL><p>") != -1)
			break;
	}
	if (strCurPath.IsEmpty())			// If path string is empty,
	{
		nPathOffset = 0;			// We're processing the root folder.
	}
	else
	{
		// Otherwise, either add or locate the current folder's path
		// and retrieve its offset.
		if (!Add_Path(strCurPath, FALSE, &nPathOffset))
		{
			Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nFailed to create internal folder item");
			return IDP_IMPORT_FAIL;
		}
	}
// Main loop processes the folder's contents one line at a time.
	while(TRUE)
	{

		if (!m_pFile->ReadString(strTemp1))			// Error if EOF reached before end of folder tag is found.
		{
			Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nUnexpected end of file");
			return IDP_IMPORT_FAIL;
		}
		strDebugMsg = strTemp1;	// Copy line and format it in case we have to report a parsing error.
		if (strDebugMsg[strDebugMsg.GetLength()-1] == '\r')
		{
			strTemp2 = strDebugMsg.Left(strDebugMsg.GetLength() -1);
			strDebugMsg = strTemp2;
		}
// Folder tag
		// (Form is "<DT><H3 misc data>Folder Name</H3>")
		if ((iPos1 = strTemp1.Find("<DT><H3")) != -1)	// Folder tag?
		{
			// If there's a Bookmark pending, write it to the file. This seems clumsy, but it's the
			// easiest way to handle multi-line comments.
			if (bGotURL)
			{
				bGotURL = FALSE;			// Reset pending URL flag.
				bInComment = FALSE;			// Reset strComment processing flag.
				if (strComment.IsEmpty())		// If no strComment, pass NULL, otherwise pass the strComment itself.
					nResult = Add_Item(nPathOffset, strPage, strURL, NULL, lAddDate, lModDate, lVisitDate);
				else
					nResult = Add_Item(nPathOffset, strPage, strURL, strComment, lAddDate, lModDate, lVisitDate);
				if (!nResult)
				{
					Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nFailed to add item to document");
					return IDP_IMPORT_FAIL;
				}
			}
			iPos1 += 7;								// Skip past "<DT><H3"
			strTemp2 = strTemp1.Mid(iPos1);					// Truncate the line.
			if ((iPos1 = strTemp2.Find('>')) == -1)		// Make sure end of <H3 tag exists.
			{
				strErrMsg.Format("CSUDoc::Parse_NS_Folder:\nHTML error in line:\n%s\nMissing '>' marker", strDebugMsg);
				Report_Fatal_Error(strErrMsg);
				return IDP_IMPORT_FAIL;
			}
			iPos1 += 1;								// Move past end of <H3 tag.
			strTemp1 = strTemp2.Mid(iPos1);
			if ((iPos1 = strTemp1.Find("</H3>")) == -1)	// Make sure </H3> end marker is present.
			{
				strErrMsg.Format("CSUDoc::Parse_NS_Folder:\nHTML error in line:\n%s\nMissing </H3> tag", strDebugMsg);
				Report_Fatal_Error(strErrMsg);
				return IDP_IMPORT_FAIL;
			}
			strTemp2 = strTemp1.Left(iPos1);				// Grab the folder name.
			if (strTemp2.IsEmpty())					// If name is blank,
			{
				++nBlankFolderCount;				// Provide default name.
				strTemp2.Format("Folder (%u)", nBlankFolderCount);
				if (strCurPath.IsEmpty())				// Format and write appropriate
					strTemp1 = "Root Folder";			// message to the error log.
				else
				{
					strTemp1 = "Folder ";
					strTemp1 += strCurPath.Left(strCurPath.GetLength() -1);
				}
				strErrMsg.Format("%s contains a folder with no name\nDefault name %s provided\n", strTemp1, strTemp2);
				m_pLogFile->WriteString(strErrMsg);
			}
			Fix_FileName(strTemp2);					// Clean up illegal characters for MSIE compatibility.
			strTemp2 += '\\';
			nResult = Parse_NS_Folder(strCurPath, strTemp2);	// Try to parse new folder.
			if (nResult)
				return nResult;		// Report error if necessary.
			continue;
		}
// Bookmark tag.
		if ((iPos1 = strTemp1.Find("<DT><A")) != -1)		// Branch if URL tag.
		{
			// If there's a Bookmark pending, write it to the file. This seems clumsy, but it's the
			// easiest way to handle multi-line comments.
			if (bGotURL)
			{
				bGotURL = FALSE;			// Reset pending URL flag.
				bInComment = FALSE;			// Reset strComment processing flag.
				if (strComment.IsEmpty())		// If no strComment, pass NULL, otherwise pass the strComment itself.
					nResult = Add_Item(nPathOffset, strPage, strURL, NULL, lAddDate, lModDate, lVisitDate);
				else
					nResult = Add_Item(nPathOffset, strPage, strURL, strComment, lAddDate, lModDate, lVisitDate);
				if (!nResult)
				{
					Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nFailed to add item to document");
					return IDP_IMPORT_FAIL;
				}
			}
			strPage.Empty();			// Reset strings for the bookmark.
			strURL.Empty();
			strComment.Empty();
			iPos1 += 6;				// Skip past the "<DT><A" marker.
			strTemp2 = strTemp1.Mid(iPos1);
			if ((iPos1 = strTemp2.Find("ADD_DATE=\"")) == -1)			// ADD_DATE tag present?
				lAddDate = 0;										// No date if so.
			else													// Otherwise,
			{
				iPos1 += 10;											// Strip off text before the value.
				strTemp1 = strTemp2.Mid(iPos1);
				lAddDate = strtoul(strTemp1, NULL, 10);				// And convert ASCII to unsigned long.
			}
			// Do the same for the "LAST_MODIFIED" and "LAST_VISIT" tags.
			if ((iPos1 = strTemp2.Find("LAST_MODIFIED=\"")) == -1)
				lModDate = 0;
			else
			{
				iPos1 += 15;
				strTemp1 = strTemp2.Mid(iPos1);
				lModDate = strtoul(strTemp1, NULL, 10);
			}
			if ((iPos1 = strTemp2.Find("LAST_VISIT=\"")) == -1)
				lVisitDate = 0;
			else
			{
				iPos1 += 12;
				strTemp1 = strTemp2.Mid(iPos1);
				lVisitDate = strtoul(strTemp1, NULL, 10);
			}
			if ((iPos1 = strTemp2.Find("HREF=\"")) == -1)		// Locate start of URL text.
			{
				Report_HTML_Err(strDebugMsg, "Missing HREF= tag");
				continue;
			}
			iPos1 += 6;										// Skip past HREF="
			strTemp1 = strTemp2.Mid(iPos1);
			if ((iPos1 = strTemp1.Find('\"')) == -1)				// Search for trailing quote.
			{
				Report_HTML_Err(strDebugMsg, "Missing close quote");
				continue;
			}
			strURL = strTemp1.Left(iPos1);							// Extract URL text.
			if (strURL.IsEmpty())								// Error if URL is blank.
			{
				Report_HTML_Err(strDebugMsg, "URL string is empty");
				continue;
			}
			if ((iPos1 = strTemp1.Find('>', iPos1)) == -1)				// Search for close of <A tag.
			{
				Report_HTML_Err(strDebugMsg, "Missing > marker");
				continue;
			}
			iPos1 += 1;										// Skip past the ">"
			strTemp2 = strTemp1.Mid(iPos1);							// Extract strPage title.
			if ((iPos1 = strTemp2.Find("</A>")) == -1)			// Search for "</A>" end marker.
			{
				Report_HTML_Err(strDebugMsg, "Missing </A> tag");
				continue;
			}
			if (iPos1 <= 50)									// Truncate strPage title to 50 chars if necessary
				strPage = strTemp2.Left(iPos1);						// for compatibility with MSIE.
			else
				strPage = strTemp2.Left(50);
			if (strPage.IsEmpty())								// If strPage title is blank,
			{
				strTemp2 = strURL;								// Create a title from the URL
				strPage = strTemp2.Left(50);
				Fix_FileName(strPage);								// Convert illegal characters (MSIE compat. again!)
				strErrMsg.Format("Bookmark for URL \"%s\" has no name\nDefault name \"%s\" provided\n", strURL, strPage);
				m_pLogFile->WriteString(strErrMsg);
				m_bErrorsLogged = TRUE;
			}
			else
				Fix_FileName(strPage);
			// Set flag to indicate that URL is pending. Note that because of the way
			// comments are stored in the file, we cannot store the URL until we have
			// parsed the next line in the file.
			bGotURL = TRUE;
			continue;				// Continue parsing.
		}
		// Comment tags are of the form "<DD> strComment text"
		// Note that comments may span multiple lines, and the end of the comment
		// does _not_ have a corresponding </DD> marker -- the comment ends when
		// the next Folder, BookMark, or End of Folder tag is reached.
// Comment tag.
		if ((iPos1 = strTemp1.Find("<DD>")) != -1)
		{
			bInComment = TRUE;			// Set flag to indicate that we're processing a comment sequence.
			iPos1 += 4;					// Skip past the <DD> marker.
			strComment = strTemp1.Mid(iPos1);	// Extract remainder of line verbatim.
			continue;					// Continue parsing.
		}
		if ((iPos1 = strTemp1.Find("</DL><p>")) != -1)	// Done processing when end of folder tag is reached.
			break;
		if (bInComment)								// If none of the above, and flag is set, we're in
		{											// the middle of a comment.
			strComment += '\n';						// Append CR/LF to existing comment.
			strComment += strTemp1;					// Append additional comment text.
		}
	}
	// If there's a URL pending, write it to the file.
	if (bGotURL)
	{
		bGotURL = FALSE;			// Reset pending URL flag.
		bInComment = FALSE;			// Reset strComment processing flag.
		if (strComment.IsEmpty())		// If no strComment, pass NULL, otherwise pass the strComment itself.
			nResult = Add_Item(nPathOffset, strPage, strURL, NULL, lAddDate, lModDate, lVisitDate);
		else
			nResult = Add_Item(nPathOffset, strPage, strURL, strComment, lAddDate, lModDate, lVisitDate);
		if (!nResult)
		{
			Report_Fatal_Error("CSUDoc::Parse_NS_Folder:\nFailed to add item to document");
			return IDP_IMPORT_FAIL;
		}
	}
	// Signal success to caller.
	return 0;
}


// Function replaces any illegal characters in the
// filename passed with the user-specified character (default is $).
// This is required for compatibility with MSIE bookmarks and folders,
// which represent actual files and folders on disk.
void CSUDoc::Fix_FileName(CString& fname)
{
	char		*buffer = new char[fname.GetLength() + 1];
	char		*pDest = buffer;
	const char	*pSource = (LPCTSTR) fname;

	while (*pSource)
	{
		switch (*pSource)
		{
		case '\\':			// Illegal filename character:
		case '/':
		case ':':
		case '?':
		case '*':
		case '\"':
		case '<':
		case '>':
		case '|':
			*pDest = m_chReplace;	// Change to user-specified character.
			break;
		case '\t':			// Tab:
			*pDest = ' ';			// Change to space.
			break;
		default:			// Any other character:
			*pDest = *pSource;			// Copy it.
		}
		++pSource;
		++pDest;
	}
	*pDest = '\0';		// Make sure string is terminated.
	fname = buffer;		// Copy modified string.
	fname.TrimLeft();	// Remove leading whitespace.
	fname.TrimRight();	// Remove trailing whitespace.
	delete [] buffer;
}

// Callback function used by qsort() library function to
// sort our list of URLs. This is a two-key sort,
// first key is the page name, second key is the URL's parent folder.
int __cdecl Compare_Items(const void * p1, const void * p2)
{
	ASSERT(p1 != NULL && p2 != NULL);

	int					result;
	const DOC_ELEMENT	*i1 = (DOC_ELEMENT*) p1;
	const DOC_ELEMENT	*i2 = (DOC_ELEMENT*) p2;

	result = i1->pPage->CompareNoCase(*i2->pPage);	// Are strings identical?
	if (result)										// If not,
		return result;								// return result to caller.
	// If strings are identical, subkey is the directory in which they reside.
	if (i1->nPath > i2->nPath)
		return 1;
	if (i1->nPath < i2->nPath)
		return -1;
	else
		return 0;
}

// Resets the "processed" flags used by recursive functions.
void CSUDoc::Reset_Path_Flags()
{
	for (UINT i = 0; i < m_nPathCount; m_pPathList[i++].bProcessed = FALSE);
}

// Function reads MSIE and/or Netscape bookmarks and
// adds them to the current document.
void CSUDoc::Do_Import()
{
	BOOL	bFatal = FALSE;

	m_bItemsAdded = m_bItemsChanged = FALSE;
	if (!Start_Import_Log())
	{
		AfxMessageBox(IDP_ERR_LOG_FAIL);
		return;
	}
	Add_Path("\\", TRUE);		// Add path for root folder.
	m_pStatDlg = new CStatDlg;
	m_pStatDlg->CenterWindow();
	m_pStatDlg->ShowWindow(SW_SHOW);
	if (m_bNSPresent)
	{
		m_pStatDlg->Show_Message("Importing Netscape 3.x bookmarks");
		m_pLogFile->WriteString("Beginning Netscape 3.x import\n");
		if (Parse_NS_File(m_strNSPath) == IDP_IMPORT_FAIL)
			bFatal = TRUE;
		if (!m_bErrorsLogged)
			m_pLogFile->WriteString("Netscape 3.x import complete ... no errors.\n\n");
	}
	if (m_bNSCPresent)
	{
		m_pStatDlg->Show_Message("Importing Netscape 4.x bookmarks");
		m_pLogFile->WriteString("Beginning Netscape 4.x import\n");
		if (Parse_NS_File(m_strNSCPath) == IDP_IMPORT_FAIL)
			bFatal = TRUE;
		if (!m_bErrorsLogged)
			m_pLogFile->WriteString("Netscape 4.x import complete ... no errors.\n\n");
	}
	if (m_bIEPresent)
	{
		m_pStatDlg->Show_Message("Importing MSIE bookmarks");
		m_pLogFile->WriteString("\nBeginning MSIE import\n\n");
		if (!Scan_Folder(NULL))
			bFatal = TRUE;
		else
			m_pLogFile->WriteString("MSIE import complete ... no errors.\n\n");
	}
	Do_Import_Cleanup();
	m_pStatDlg->DestroyWindow();
	delete m_pStatDlg;
	m_pStatDlg = NULL;
	End_Import_Log();
	if (m_bErrorsLogged)
		Query_Show_Log(bFatal);
	SetModifiedFlag(m_bItemsChanged || m_bItemsAdded);
}


// Creates MSIE bookmarks from the current document.
// This function is recursive, so it is called once
// with the path_offset parm set to 0, indicating to start with
// the root folder. It then calls itself with the offset of each
// child folder until all folders have been processed.
// Returns FALSE and stops recursion only if a folder cannot be created.

// Function works by recursing through the tree starting at the root.
// If a folder does not exist, it is created.
// If a folder does exist, any existing .URL files in that folder are removed, if possible
//  (function does not attempt to remove or alter existing read-only files).
// Process repeats until the end of a branch is reached and all folders
// along the way have been either created or cleaned. The individual bookmarks (.url files)
// within each folder are created as the function walks back up the tree.
// Failure to create a bookmark (e.g., read-only file, invalid URL syntax, etc.)
// presents a warning message but does not stop the process.

BOOL CSUDoc::Export_MSIE_Path(UINT nPathOffset)
{
	int			iResult,
				iErrCode;
	UINT		i,
				nChild,
				nSelected;
	CString		strMessage,
				strPath1,
				strPath2;

	nSelected = m_pPathList[nPathOffset].nSelected;	// Get folder's selection state.
	m_pPathList[nPathOffset].bProcessed = TRUE;		// Mark folder as processed.
	strPath1 = m_strIEPath;							// Retrieve path to "Favorites" folder.
	if (nPathOffset != 0)							// If not the root,
		strPath1 += *m_pPathList[nPathOffset].pPath;	// append relative path to "Favorites" path.
	strPath2 = strPath1.Left(strPath1.GetLength() - 1);	// Strip trailing backslash.

// This section of code was updated for version 1.4.
// Clean_Folder() is now called once, before any folders are exported,
// and wipes the entire Favorites folder tree. Therefore,
// this function no longer has to worry about "cleaning" any folders individually.
// Also, any folders that _do_ exist must contain "untouchable" files,
// and therefore should be locked.

	if (nSelected != 1)								// If folder is selected,
	{
		strMessage.Format("Creating folder %s", strPath2);	// Tell user what we're doing.
		m_pStatDlg->Show_Message(strMessage);
		iResult = _mkdir(strPath2);			// Try to create the folder.
		iErrCode = errno;					// Save possible error code.
		if (iResult == -1)					// Did creation fail?
		{
			if (iErrCode != EEXIST)			// Yes ... if the folder didn't already exist,
			{
				strMessage.Format("Could not create folder:\n\"%s\"\nError code = %08X", strPath2, iErrCode);
				AfxMessageBox(strMessage, MB_ICONEXCLAMATION);
				return FALSE;				// Report failure and quit.
			}
			else	// Creation failed because folder already exists, so 
			{
				m_pPathList[nPathOffset].bLocked = TRUE;	// mark it as locked.
			}
		}
		else		// If we just created the folder, it won't contain anything besides bookmarks, so
		{
			m_pPathList[nPathOffset].bLocked = FALSE;	// make sure it's not locked.
		}
	}
	// Process any child folders first.
	while(Has_Children(nPathOffset, &nChild))
	{
		if (!Export_MSIE_Path(nChild))
			return FALSE;
	}
	// Let user know what's going on.
	strMessage.Format("PROCESSING: %s", strPath2);
	m_pStatDlg->Show_Message(strMessage);
	// If this folder is selected,
	if (nSelected != 1)
	{
		// Process the bookmarks it contains.
		for (i = 0; i < m_nItemCount; i++)
		{
			if (m_pDocElements[i].nPath == nPathOffset)
				if (!Create_URL_File(strPath1, i))
					return FALSE;
		}
	}
	// If folder is not selected,
	else
	{
		if (nPathOffset != 0)								// If folder is not the root,
		{
			if (_rmdir(strPath2) == 0)						// try to remove it and
				m_pPathList[nPathOffset].bLocked = FALSE;	// mark it as unlocked if successful.
		}
		else												// Folder is the root,
			m_pPathList[nPathOffset].bLocked = TRUE;			// so lock it.
	}
	return TRUE;
}

// Called by Export_MSIE_Folder()
// Creates an "Internet Shortcut" file in the specified directory.
// INPUT:
//    szPath = fully qualified path (including trailing backslash) to the
//				folder in which file is to be created.
//    nItemOffset = offset into the document array for the specified bookmark.
// RETURNS:
//    FALSE if unable to access OLE interface, otherwise TRUE.
// Also displays appropriate error message if unable to create the file
// as requested.
BOOL CSUDoc::Create_URL_File(LPCTSTR szPath, UINT nItemOffset)
{
	CString					strMessage,
							strFileName;
	HRESULT					hResult;
	IPersistFile			*pIPF = NULL;
	IUniformResourceLocator *pIURL = NULL;

	// Try to instantiate an IUniformResourceLocator object.
	hResult = CoCreateInstance(CLSID_InternetShortcut,
		NULL, CLSCTX_INPROC_SERVER, IID_IUniformResourceLocator, (void **) &pIURL); 
	// If that worked, try to get pointer to IUniformResourceLocator's
	// IPersistFile interface.
	if (SUCCEEDED(hResult))
	{
		hResult = pIURL->QueryInterface(IID_IPersistFile, (void **) &pIPF);
		if (!SUCCEEDED(hResult))
			pIURL->Release();
	}
	// If either pointer is NULL, give appropriate message and return FALSE.
	if (pIURL == NULL || pIPF == NULL)
	{
		AfxMessageBox("OLE Init failed", MB_ICONINFORMATION);
		return FALSE;
	}
	// Create fully-qualified filename.
	strFileName.Format("%s%s.url", szPath, *m_pDocElements[nItemOffset].pPage);
	// Covert string to OLE format.
	MultiByteToWideChar(CP_ACP, 0, strFileName, -1, szWide, _MAX_PATH);
	// Set OLE object's URL.
	if (pIURL->SetURL(*m_pDocElements[nItemOffset].pUrl, IURL_SETURL_FL_GUESS_PROTOCOL) == S_OK)
	{
		// If URL was set, save the object to the file.
		if (pIPF->Save(szWide, TRUE) == S_OK)
		{
			pIPF->SaveCompleted(szWide);
		}
		else
		{
			// Error if save failed.
			strMessage.Format("File creation failed.\n\nFilename: %s", strFileName);
			AfxMessageBox(strMessage, MB_ICONINFORMATION);
		}
	}
	else
	{
		strMessage.Format("Invalid URL syntax.\n\nPage: %s\nURL:%s",
							*m_pDocElements[nItemOffset].pPage,
							*m_pDocElements[nItemOffset].pUrl);
		AfxMessageBox(strMessage, MB_ICONINFORMATION);
	}
	// Release OLE objects.
	pIPF->Release();
	pIURL->Release();
	return TRUE;
}

// Called by Do_MSIE_Export
// Prepares the specified directory by deleting all .URL files and subfolders
// it contains. However, subfolders will not be removed unless they are completely
// empty after any .URL files they contain have been deleted.
// Function expects a fully-qualified path string with trailing backslash.
//
// Note that this function changed in version 1.4.
// In previous versions, it was called by Export_MSIE_Folder, as each folder
// was exported, and simply removed any .URL files from the specified folder.
// Now it is called only once by Do_MSIE_Export, and uses recursion to zap the
// entire "Favorites" folder tree. This change eliminates the need to "lock"
// any MSIE folders except for those that contain non-.URL files.
//
void CSUDoc::Clean_Folder(LPCTSTR pszPath)
{
	HANDLE			hFind;
	CString			strSearchSpec,
					strFileSpec;
	WIN32_FIND_DATA fd;

	strSearchSpec = pszPath;
	strSearchSpec += "*.*";		// Search for all files.
	if ((hFind = FindFirstFile(strSearchSpec, &fd)) != INVALID_HANDLE_VALUE)
	{
		do
		{
			if ((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)	// Not a directory?
			{
				strFileSpec = pszPath;					// Get fully-qualified filename.
				strFileSpec += fd.cFileName;
				strFileSpec.MakeUpper();				// Force to uppercase.
				if (strFileSpec.Right(4) == ".URL")		// Is the extension ".URL"?
				{
					try {								// Yes -- try to delete the file.
						CFile::Remove(strFileSpec);
					}
					catch (CFileException* e)
					{
						e->Delete();
					}
				}
			}
			else										// Object is a directory.
			{
				if (strcmp(fd.cFileName, ".") != 0)			// If it's not "."
				{
					if (strcmp(fd.cFileName, "..") != 0)		// or ".."
					{
						strFileSpec.Format("%s%s\\", pszPath, fd.cFileName);	// Get its fully-qualified name.
						Clean_Folder(strFileSpec);								// Clean it.
						_rmdir(strFileSpec);									// Then try to remove it.
					}
				}
			}
		} while (FindNextFile(hFind, &fd));
		FindClose(hFind);
	}
}

// Called by ID_EXPORT commands.
// Prepares for, then executes the MSIE export function.
void CSUDoc::Do_MSIE_Export()
{
	BOOL bResult;

	m_pStatDlg = new CStatDlg;		// Create new Status dialog object.
	if (m_pStatDlg == NULL)			// Report error and bail out if dialog creation failed.
	{
		AfxMessageBox("Internal Error:\nCStatDlg creation failed in CSUDoc::Do_MSIE_Export()"
						"\nOperation cancelled -- existing MSIE folders and"
						"\nbookmarks remain unchanged.", MB_ICONINFORMATION);
		return;
	}
	m_pStatDlg->CenterWindow();		// Center the dialog.
	for (UINT i = 0; i < m_nPathCount; i++)	// Reset processing flags.
		m_pPathList[i].bProcessed = FALSE;
	Clean_Folder(m_strIEPath);
	bResult = Export_MSIE_Path(0);	// Do the export, starting with the root folder.
	m_pStatDlg->DestroyWindow();	// Close the dialog.
	delete m_pStatDlg;				// Delete the dialog object.
	m_pStatDlg = NULL;				// NULL object pointer.
	UpdateAllViews(NULL);			// Refresh the tree control.
	if (!bResult)					// Indicate final status to user.
		AfxMessageBox("MSIE Export failed.", MB_ICONEXCLAMATION);
}

// Deletes the specified item from the document.
void CSUDoc::Delete_Item(UINT nOffset)
{
	void			*p1,
					*p2;
	size_t			size = sizeof (DOC_ELEMENT);
	DOC_ELEMENT		*pNewArray;

	// Delete CStrings belonging to the item, if they exist.
	if (m_pDocElements[nOffset].pPage)
		delete m_pDocElements[nOffset].pPage;
	if (m_pDocElements[nOffset].pUrl)
		delete m_pDocElements[nOffset].pUrl;
	if (m_pDocElements[nOffset].pComment)
		delete m_pDocElements[nOffset].pComment;
	// Adjust the item count.
	--m_nItemCount;
	// Mark document as dirty.
	SetModifiedFlag();
	if (m_nItemCount == 0)		// No items left?
	{
		delete [] m_pDocElements;	// Delete the entire array,
		m_pDocElements = NULL;		// NULL the pointer,
		m_nDocArraySize = 0;	// set its size to zero,
		return;						// and we're done.
	}
	// If the deleted item was not the last one in the array,
	if (nOffset < m_nItemCount)
	{
		// Close the gap left by the deleted item
		p1 = &m_pDocElements[nOffset];
		p2 = &m_pDocElements[nOffset + 1];
		memmove(p1, p2, (m_nItemCount - nOffset) * size);
	}
	// If array has shrunk by the specified increment,
	if ((m_nDocArraySize - m_nItemCount) == DOC_ARRAY_INCREMENT)
	{
		// Set new array size.
		m_nDocArraySize -= DOC_ARRAY_INCREMENT;
		// Reallocate the array.
		pNewArray = (DOC_ELEMENT*) realloc(m_pDocElements, m_nDocArraySize * size);
		m_pDocElements = pNewArray;
	}
}

// Function takes a path string (with trailing backslash)
// and replaces it with the name of the last folder in the path.
void CSUDoc::Friendly_Name(CString & strPath)
{
	int		iPos;
	CString strScratch;

	if (strPath.GetLength() == 1)
	{
		strPath = "Root Folder";
		return;
	}
	strScratch = strPath.Left(strPath.GetLength() - 1);
	iPos = strScratch.ReverseFind('\\') + 1;
	if (iPos)
		strPath = strScratch.Mid(iPos);
	else
		strPath = strScratch;
}

// Moves specified source folder and its children
// to the specified destination folder. Note that folders are
// specified by their offsets into the m_pPathList array.
// Returns FALSE if folder cannot be moved because:
//   1. Folder with same name already exists as a child of the destination.
//   2. Destination is a child of the source folder.
//   3. Resulting new paths would be too long.
BOOL CSUDoc::Move_Folder(UINT source, UINT dest)
{
	int			iBaseLen,
				iSourceLen,
				iTargLen;
	BOOL		bAll,
				bNone;
	UINT		i,
				nCurrentPath,
				nParent,
				nOldParent;
	CString		strSource = *m_pPathList[source].pPath,
				strDest = *m_pPathList[dest].pPath,
				strScratch1,
				strScratch2;

	nOldParent = m_pPathList[source].nParent;		// Get parent of folder that moved.
	iBaseLen = m_strIEPath.GetLength();				// Get length of path to the "Favorites" folder
	if (dest != 0)									// If destination is not the root,
		iBaseLen += strDest.GetLength();			// Adjust length of destination path.
	if (nOldParent == 0)							// Get length of parent path for source folder.
		iSourceLen = 0;
	else
		iSourceLen = m_pPathList[m_pPathList[source].nParent].pPath->GetLength();
	for (i = 1; i < m_nPathCount; i++)			// Loop through the paths
	{
		if (m_pPathList[i].pPath->Find(strSource) == 0)					// If it will be moved,
		{
			iTargLen = m_pPathList[i].pPath->GetLength() - iSourceLen;	// make sure resulting path
			if ((iTargLen + iBaseLen) > 200)							// won't be too long.
				return FALSE;											// Can't move if it is.
		}
	}
	strScratch1 = strSource;					// Get source path.
	Friendly_Name(strScratch1);					// Extract last folder name.
	for (i=1; i < m_nPathCount; i++)			// Walk through the paths again.
	{
		if (m_pPathList[i].nParent == dest)		// Is path a child of the destination?
		{
			strScratch2 = *m_pPathList[i].pPath;					// Grab the path.
			Friendly_Name(strScratch2);								// Extract last dir in the path.
			if (strScratch1.CompareNoCase(strScratch2) == 0)		// If name is the same,
				return FALSE;										// can't move it.
		}
	}
	for (i=1; i < m_nPathCount; i++)			// Loop through the paths yet again.
	{
		if (m_pPathList[i].pPath->Find(strSource) == 0)		// Does it belong to the source folder?
		{
			if (dest != 0)									// Build new name if so.
				strScratch1 = strDest + m_pPathList[i].pPath->Mid(iSourceLen);
			else
				strScratch1 = m_pPathList[i].pPath->Mid(iSourceLen);
			*m_pPathList[i].pPath = strScratch1;			// Install the new name.
			if (i == source)								// If it's the source path,
				m_pPathList[i].nParent = dest;				// set its new nParent offset.
		}
	}
	// Update the selection state.
	Reset_Path_Flags();
	// Handle source folder first (folder can't be moved to its own children).
	if (Has_Children(nOldParent, &nCurrentPath))
	{
		bAll = TRUE;								// Flag indicates all subfolders are selected.
		bNone = TRUE;								// Flag indicates no subfolders are selected.
		for (i = 1; i < m_nPathCount; i++)		// Walk the path array.
		{	
			if (m_pPathList[i].nParent != nOldParent)	// Skip if not a child of the source folder.
				continue;
			if (m_pPathList[i].nSelected != 2)			// If not fully selected,
				bAll = FALSE;							// then not all children are selected.
			if (m_pPathList[i].nSelected != 1)			// If fully or partially selected,
				bNone = FALSE;							// then at least some of the children are selected.
		}
		if (bNone)										// No children selected at all?
			m_pPathList[nOldParent].nSelected = 1;		// Then it's not selected either.
		else
		{
			if (bAll)									// All children fully selected?
				m_pPathList[nOldParent].nSelected = 2;	// Then so is parent.
			else
				m_pPathList[nOldParent].nSelected = 3;	// Otherwise, parent is partially selected.
		}
	}
	else												// Old source folder has no remaining children,
	{
		if (m_pPathList[nOldParent].nSelected == 3)		// so if it was partially selected,
			m_pPathList[nOldParent].nSelected = 2;		// it is now fully selected.
		// (If it wasn't selected before the move, it remains unselected!)
	}
	// Now handle the destination folder.
	nCurrentPath = source;
	do													// Loop works from source folder upwards.
	{
		nParent = m_pPathList[nCurrentPath].nParent;	// Get parent of current folder.
		bAll = TRUE;									// Reset flags.
		bNone = TRUE;
		for (i = 1; i < m_nPathCount; i++)			// Same algorithm as previous loop.
		{
			if (m_pPathList[i].nParent != nParent)
				continue;
			if (m_pPathList[i].nSelected != 2)
				bAll = FALSE;
			if (m_pPathList[i].nSelected != 1)
				bNone = FALSE;
		}
		// Slight difference here; since we're working from the bottom up,
		// we know that the target folder has children.
		if (bNone)
			m_pPathList[nParent].nSelected = 1;
		else
		{
			if (bAll)
				m_pPathList[nParent].nSelected = 2;
			else
				m_pPathList[nParent].nSelected = 3;
		}
		nCurrentPath = nParent;	// Move up a level.
	} while (nCurrentPath);		// Done once we reach the root.
	return TRUE;		// Success!
}

// Function resolves duplicate bookmark names
// Specifically, for bookmarks in the same folder with the same name but
// different URLs). The import process automatically discards bookmarks
// that are true duplicates, with same name _and_ URL.
// Returns TRUE if any bookmarks were renamed.
BOOL CSUDoc::Resolve_Dupes()
{
	int			iLen1,
				iLen2;
	BOOL		bFound = FALSE,
				bResult = FALSE,
				bValid;
	UINT		i,
				j,
				k,
				l,
				nCount,
				nDupCount,
				nAnchor;
	CString		*scratch,
				s1,
				s2,
				s3,
				msg;
	DUP_ELEMENT	*pDupes;

	m_pLogFile->WriteString("\nResolving duplicate bookmarks:\n\n");
	// Make array same size as the document.
	pDupes = new DUP_ELEMENT[m_nItemCount];
	ASSERT(pDupes!= NULL);

	// Outer loop walks the folder array.
	for (i = 0; i < m_nPathCount; i++)
	{
		nCount = 0;								// Reset element count.
		if (i == 0)
			msg = "Processing Root Folder\n";
		else
			msg.Format("Processing %s\n", *m_pPathList[i].pPath);
		m_pLogFile->WriteString(msg);
		for (j = 0; j < m_nItemCount; j++)		// Walk the document.
		{
			if (m_pDocElements[j].nPath != i)	// Bookmark in current folder?
				continue;
			scratch = new CString(*m_pDocElements[j].pPage);	// Copy its name.
			ASSERT (scratch != NULL);
			pDupes[nCount].pName = scratch;		// and add it to the array.
			pDupes[nCount].nOffset = j;
			++nCount;
		}
		bFound = FALSE;
		if (nCount > 1)							// Nothing to do if folder has only one bookmark!
		{
			// Sort the names into alphabetical order.
			qsort((void*) pDupes, nCount, sizeof(DUP_ELEMENT), Compare_Dupes);
			// Walk through the array.
			for (j = 0; j < nCount -1; j++)
			{
				// Two elements with same name?
				if (pDupes[j].pName->CompareNoCase(*pDupes[j + 1].pName) == 0)
				{
					bFound = TRUE;		// Set flag.
					nDupCount = 1;		// Initialize count for numeric suffixes.
					nAnchor = j;		// Save position of first element.
					// Walk through the remainder of the array.
					for (k = nAnchor + 1; k < nCount; ++nDupCount)
					{
						s1 = *pDupes[k].pName;			// Copy folder name.
						iLen1 = s1.GetLength();			// Get its length.
						s2.Format(" (%u)", nDupCount);	// Create numeric suffix.
						iLen2 = s2.GetLength();			// Get its length.
						if ((iLen1 + iLen2) < 50)		// If it's not too long,
							s1 += s2;					// just append the suffix.
						else
						{
							s3 = s1.Left(50 - iLen2);	// Otherwise chop the name
							s1 = s3 + s2;				// then append the suffix.
						}
						bValid = TRUE;					// Set flag.
						for (l = 0; l < nCount; l++)	// Make sure new name doesn't already exist!
						{
							if (s1.CompareNoCase(*pDupes[l].pName) == 0)
							{
								bValid = FALSE;
								break;
							}
						}
						if (!bValid)					// If new name already exists,
							continue;					// Try again with next suffix number.
						// Write message to the log.
						msg.Format("  Duplicate  \"%s\"\n"
								   "  renamed to \"%s\"\n\n", *pDupes[k].pName, s1);
						m_pLogFile->WriteString(msg);
						m_bErrorsLogged = TRUE;
						// Install the new name.
						*pDupes[k].pName = s1;
						// Next element
						++k;
						// Done if name doesn't match anchor name.
						if (k < nCount)
							if (pDupes[nAnchor].pName->CompareNoCase(*pDupes[k].pName) != 0)
								break;
					}
					j = k - 1;		// Ajust array position and keep going.
				}
			}
		}
		if (bFound)					// If any dupes were found,
		{
			bResult = TRUE;			// Set flag.
			for (j = 0; j < nCount; j++)	// Then copy the new names back into the document.
				*m_pDocElements[pDupes[j].nOffset].pPage = *pDupes[j].pName;
		}
		for (j = 0; j < nCount; j++)	// Clean out work array.
			delete pDupes[j].pName;
	}	// Continue with next folder.
	delete [] pDupes;		// Delete the work array.
	// Write string to log if no dupes were bFound.
	if (!bResult)
		m_pLogFile->WriteString("\nNo duplicates bFound.\n");
	else
		m_bItemsChanged = TRUE;
	return bResult;
}

// Called by CSUView when a folder is renamed.
// Input: offset = Offset to the path whose name is changing
//        new_name = New "Friendly" folder name (without trailing backslash)
// Returns FALSE if new name would cause any affected paths
// to be too long.
// Otherwise, adjusts all affected path strings and returns TRUE.
BOOL CSUDoc::Rename_Folder(UINT offset, LPCTSTR new_name)
{
	int			pos,
				len,
				diff;
	UINT		i;
	CString		s1,
				s2,
				s3;

	s1 = *m_pPathList[offset].pPath;			// Get current folder path.
	len = s1.GetLength();					// And its length.
	s2 = s1;								// Extract plain folder name.
	Friendly_Name(s2);						
	diff = strlen(new_name) - s2.GetLength();	// Get difference in length.
	if (diff > 0)							// If new name is longer,
	{										// check resulting path lengths.
		for (i = 1; i < m_nPathCount; i++)
		{
			if (m_pPathList[i].pPath->Find(s1) == 0)
			{
				if ((m_iBasePathLen + m_pPathList[i].pPath->GetLength() + diff) > 200)
					return FALSE;
			}
		}
	}
	if (m_pPathList[offset].nParent != 0)	// Is it not right off the root?
	{
		s2 = s1.Left(s1.GetLength() - 1);	// Strip trailing backslash
		pos = s2.ReverseFind('\\') + 1;		// Find last backslash.
		s3 = s1.Left(pos);					// Extract parent folder name.
		s3 += new_name;						// Append new folder name.
	}
	else									// Otherwise,
	{
		s3 = new_name;						// just set new folder name.
	}
	s3 += '\\';								// Append trailing backslash
	for (i = 1; i < m_nPathCount; i++)	// Walk the array.
	{
		if (i == offset)					// Target folder?
			*m_pPathList[i].pPath = s3;		// Just install the new name.
		else
		{
			if (m_pPathList[i].pPath->Find(s1) == 0)			// Child folder?
			{
				s2 = s3 + m_pPathList[i].pPath->Mid(len);	// Strip old parent, install new parent.
				*m_pPathList[i].pPath = s2;
			}
		}
	}
	SetModifiedFlag();		// Mark document at dirty.
	return TRUE;			// Success!
}

// Warns user before export if root folder is not selected.
BOOL CSUDoc::Do_Export_Warning()
{
	if (m_pPathList[0].nSelected == 1)
		return AfxMessageBox(IDP_WARN_EXPORT, MB_YESNO | MB_ICONQUESTION) == IDYES;
	else
		return TRUE;
}

// Writes non-fatal HTML error to log file.
void CSUDoc::Report_HTML_Err(CString & html, LPCTSTR err)
{
	CString msg;

	msg.Format("HTML Error in:\n%s\n%s. The line was skipped\n", html, err);
	m_pLogFile->WriteString(msg);
	m_bErrorsLogged = TRUE;
}

// Writes fatal error message to log file.
void CSUDoc::Report_Fatal_Error(LPCTSTR text)
{
	CString msg;

	msg.Format("Critical error:\n%s\nProcessing aborted\n", text);
	m_pLogFile->WriteString(msg);
	m_bErrorsLogged = TRUE;
}

// Deletes the path specified by the offset parm.
// Note that the UI ensures that this function will only
// be called if the target folder has no children, either
// folders or bookmarks, greatly simplifying the task.
void CSUDoc::Delete_Folder(UINT offset)
{
	UINT			i;
	size_t			size = sizeof(PATH_ELEMENT);
	PATH_ELEMENT	*temp,
					*p1,
					*p2;

	delete m_pPathList[offset].pPath;		// Delete path string.
	--m_nPathCount;						// Adjust path count.
	if (offset < m_nPathCount)			// If deleted path was not the last one in the array,
	{
		p1 = &m_pPathList[offset];			// Close the gap in the array.
		p2 = &m_pPathList[offset + 1];
		memmove(p1, p2, (m_nPathCount - offset) * size);
	}
	// Reallocate the array.
	temp = (PATH_ELEMENT*) realloc(m_pPathList, size * m_nPathCount);
	ASSERT(temp != NULL);
	m_pPathList = temp;
	// Adjust parent offsets for path items.
	for (i = 0; i < m_nPathCount; i++)
	{
		if (m_pPathList[i].nParent >= offset)
			--m_pPathList[i].nParent;
	}
	// Adjust path offsets for bookmark items.
	for (i = 0; i < m_nItemCount; i++)
	{
		if (m_pDocElements[i].nPath >= offset)
			--m_pDocElements[i].nPath;
	}
}

// Initializes the import log file.
// Returns FALSE if file could not be created.
BOOL CSUDoc::Start_Import_Log()
{
	CTime		time = CTime::GetCurrentTime();
	CString		strDate;

	// Format file header.
	strDate.Format("SyncURLs import log created %s\n\n", time.Format("%c"));
	// Create the file object.
	m_pLogFile = new CStdioFile;
	if (m_pLogFile != NULL)	// If successful,
	{
		// Create new import log,
		if (m_pLogFile->Open(m_strErrLog, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyWrite, NULL))
		{
			m_bErrorsLogged = FALSE;
			m_pLogFile->WriteString(strDate);
			return TRUE;
		}
		else
		{
			delete m_pLogFile;
			m_pLogFile = NULL;
		}
	}
	return FALSE;
}

// Closes the import log.
void CSUDoc::End_Import_Log()
{
	m_pLogFile->Close();
	delete m_pLogFile;
	m_pLogFile = NULL;
}

// Called after any import. Sorts the document and resolves duplicate names.
void CSUDoc::Do_Import_Cleanup()
{
	BOOL bResult;

	m_pStatDlg->Show_Message("Sorting...");
	qsort((void*) m_pDocElements, m_nItemCount, sizeof(DOC_ELEMENT), Compare_Items);
	m_pStatDlg->Show_Message("Resolving duplicate names...");
	bResult = Resolve_Dupes();
	if (bResult)
	{
		m_pStatDlg->Show_Message("Duplicate names resolved. Re-sorting...");
		qsort((void*) m_pDocElements, m_nItemCount, sizeof(DOC_ELEMENT), Compare_Items);
	}
}

// Presents query if errors encountered during import.
void CSUDoc::Query_Show_Log(BOOL bFatal)
{
	CString strMessage;

	if (bFatal)
		strMessage = "One or more critical errors occurred during import.";
	else
		strMessage =	"SyncURLs had to make some adjustments\n"
						"while importing your bookmarks.";
	strMessage += "\n\nTo see the Import Log, use the\n"
				  "\"Import Log\" command from the View Menu.";
	AfxMessageBox(strMessage, MB_ICONINFORMATION);
}

// Determine which browser(s) are present on the local system.
void CSUDoc::Detect_Browsers()
{
	char					buff2[_MAX_PATH];
	unsigned char			buff1[_MAX_PATH];
	HKEY					hKey;
	DWORD					dwLen;
	DWORD					dwType;
	HRESULT					hResult;
	IUniformResourceLocator	*pIURL = NULL;

	// Detect MSIE
	// Try to open the "User Shell Folders" key.
	hResult = RegOpenKeyEx(HKEY_CURRENT_USER, sz_IEKey, 0, KEY_QUERY_VALUE, &hKey);
	if (hResult == ERROR_SUCCESS)
	{
		// Try to read the "Favorites" data value.
		dwLen = _MAX_PATH;
		hResult = RegQueryValueEx(hKey, sz_IEData, 0, &dwType, buff1, &dwLen);
		RegCloseKey(hKey);
	}
	// If that failed,
	if (hResult != ERROR_SUCCESS)
	{
		//  try to open the "Shell Folders" key.
		hResult = RegOpenKeyEx(HKEY_CURRENT_USER, sz_IEAltKey, 0, KEY_QUERY_VALUE, &hKey);
		if (hResult == ERROR_SUCCESS)	// If successful,
		{
			// Try to read the "Favorites" data value.
			dwLen = _MAX_PATH;
			hResult = RegQueryValueEx(hKey, sz_IEData, 0, &dwType, buff1, &dwLen);
			RegCloseKey(hKey);
		}
	}
	// If value was retrieved,
	if (hResult == ERROR_SUCCESS)
	{
		// Check to see if e-vars need to be expanded.
		// (Only happens under NT -- Win 95 doesn't support the REG_EXPAND_SZ data type)
		if (dwType == REG_EXPAND_SZ)
		{
			// If so, expand e-vars, then set the value.
			ExpandEnvironmentStrings((const char *)buff1, buff2, _MAX_PATH);
			m_strIEPath = buff2;
		}
		else
			// Otherwise, just set the value.
			m_strIEPath = buff1;
		// Make sure the string points to an existing directory.
		m_bIEPresent = Validate_Path(m_strIEPath);
		if (m_bIEPresent)				// If it does, append trailing backslash if needed.
		{
			if (m_strIEPath.Right(1) != "\\")
				m_strIEPath += "\\";
		}
		else
			m_strIEPath.Empty();			// Otherwise make sure it's empty.
	}
	// If we found the "Favorites" folder, now make sure the
	// necessary OLE interface exists.
	if (m_bIEPresent)
	{
		// Try to instantiate an "Internet Shortcut" object.
		hResult = CoCreateInstance(CLSID_InternetShortcut,
									NULL,
									CLSCTX_INPROC_SERVER,
									IID_IUniformResourceLocator,
									(void **) &pIURL);
		if (SUCCEEDED(hResult))			// If successful,
			pIURL->Release();			// just release the object.
		else							// Otherwise,
		{
			m_bIEPresent = FALSE;		// disable MSIE-related functions
			m_strIEPath.Empty();		// and erase the path information.
		}
	}
	// Get length of the path to the "Favorites" folder.
	m_iBasePathLen = m_strIEPath.GetLength();
	// Detect Netscape

	if (RegOpenKeyEx(HKEY_CURRENT_USER, sz_NetscapeKey, 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS)
	{
		dwLen = _MAX_PATH;
		if (RegQueryValueEx(hKey, sz_NetscapeData, 0, &dwType, buff1, &dwLen) == ERROR_SUCCESS)
		{
			m_bNSPresent = TRUE;
			if (dwType == REG_EXPAND_SZ)
			{
				ExpandEnvironmentStrings((const char *)buff1, buff2, _MAX_PATH);
				m_strNSPath = buff2;
			}
			else
				m_strNSPath = buff1;
		}
		RegCloseKey(hKey);
	}


	// Detect Communicator
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, sz_CommunicatorKey, 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS)
	{
		m_bNSCPresent = TRUE;
		RegCloseKey(hKey);
	}
	else
		m_bNSCPresent = FALSE;
	// If Communicator key is found, create and launch the "Select User Profile" dialog.
	// The dialog takes care of retrieving individual Netscape 4.x User Profile information and,
	// if there is more than one such profile, letting the user select which one he wants to
	// work with, then sets our m_strNSCUserName and m_strNSCPath variables according.
	if (m_bNSCPresent)
	{
		CBrowsDlg dlg;			// Create the dialog,
		dlg.m_pMyDoc = this;	// give it a pointer so it can set our member variables,
		dlg.DoModal();			// and launch it.
	}
	if (m_bNSCPresent && m_bNSPresent)
	{
		if (m_strNSPath.CompareNoCase(m_strNSCPath) == 0)
		{
			m_bNSPresent = FALSE;
			m_strNSCUserName += " [shared]";
		}
	}
}

// Determines wheter or not the specified path
// refers to an existing directory.
BOOL CSUDoc::Validate_Path(LPCTSTR pszPath)
{
	HANDLE				hFind;
	WIN32_FIND_DATA		fd;

	hFind = FindFirstFile(pszPath, &fd);
	if (hFind == INVALID_HANDLE_VALUE)
		return FALSE;
	FindClose(hFind);
	return ((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0);
}

// Callback function used by qsort() in Resolve_Dupes()
int __cdecl Compare_Dupes(const void * p1, const void * p2)
{
	ASSERT(p1 != NULL && p2 != NULL);

	const DUP_ELEMENT* i1 = (DUP_ELEMENT*) p1;
	const DUP_ELEMENT* i2 = (DUP_ELEMENT*) p2;

	return (i1->pName->CompareNoCase(*i2->pName));

}

