// SUDoc.h : interface of the CSUDoc class
//
//
// SyncURLs Version 1.4
// Copyright  1998 Ziff-Davis Inc.
// First published in PC Magazine, US Edition
// January 6, 1998
// Written by Gregory A. Wolking
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_SUDOC_H__A03AD40C_1CC5_11D1_84C6_444553540000__INCLUDED_)
#define AFX_SUDOC_H__A03AD40C_1CC5_11D1_84C6_444553540000__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

// String constants.
const char sz_IEKey[] = "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\User Shell Folders";
const char sz_IEAltKey[] = "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";
const char sz_IEData[] = "Favorites";
const char sz_NetscapeKey[] = "Software\\Netscape\\Netscape Navigator\\Bookmark List";
const char sz_NetscapeData[] = "File Location";
const char sz_CommunicatorKey[] = "SOFTWARE\\Netscape\\Netscape Navigator\\Users";

int __cdecl Compare_Items(const void* p1, const void* p2);
int __cdecl Compare_Dupes(const void* p1, const void* p2);

class CSUDoc : public CDocument
{
protected: // create from serialization only
	CSUDoc();
	DECLARE_DYNCREATE(CSUDoc)

// Attributes
public:

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CSUDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void Serialize(CArchive& ar);
	virtual BOOL OnOpenDocument(LPCTSTR lpszPathName);
	virtual void DeleteContents();
	//}}AFX_VIRTUAL

// Implementation
public:
	// Variables
	int				m_iBasePathLen;
	UINT			m_nItemCount;
	UINT			m_nPathCount;
	BOOL			m_bNSCPresent;
	BOOL			m_bNSPresent;
	PATH_ELEMENT *	m_pPathList;
	DOC_ELEMENT *	m_pDocElements;
	CString			m_strNSPath;
	CString			m_strNSPageTitle;
	CString			m_strNSCPageTitle;
	CString			m_strNSCPath;
	CString			m_strNSCUserName;
	// Functions
	void Do_MSIE_Export();
	void Fix_FileName(CString& fname);
	void Friendly_Name(CString& strPath);
	void Delete_Item(UINT offset);
	void Reset_Path_Flags();
	void Delete_Folder(UINT offset);
	BOOL Rename_Folder(UINT offset, LPCTSTR new_name);
	BOOL Move_Folder(UINT source, UINT dest);
	BOOL Add_Item(UINT nPathOffset, LPCTSTR lpszPage, LPCTSTR lpszURL, LPCTSTR lpszComment, LONG nAdd = 0, LONG nMod = 0, LONG nVisit = 0, UINT *pnOffset = NULL);
	BOOL Has_Children(UINT nOffset, UINT* pnChild);
	BOOL Add_Path(LPCTSTR lpszPath, BOOL bLocked, UINT* pnOffset = NULL);
	BOOL Do_Export_Warning(void);
	virtual ~CSUDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	BOOL m_bOpenSpecial;
	char m_chReplace;
	UINT m_nDocArraySize;
	//{{AFX_MSG(CSUDoc)
	afx_msg void OnUpdateFileSave(CCmdUI* pCmdUI);
	afx_msg void OnUpdateFileSaveAs(CCmdUI* pCmdUI);
	afx_msg void OnUpdateExportNs(CCmdUI* pCmdUI);
	afx_msg void OnExportText();
	afx_msg void OnUpdateExportText(CCmdUI* pCmdUI);
	afx_msg void OnUpdateExportMsie(CCmdUI* pCmdUI);
	afx_msg void OnUpdateExportBoth(CCmdUI* pCmdUI);
	afx_msg void OnFileManual();
	afx_msg void OnUpdateExportManual(CCmdUI* pCmdUI);
	afx_msg void OnUpdateExportNsc(CCmdUI* pCmdUI);
	afx_msg void OnViewLog();
	afx_msg void OnUpdateViewLog(CCmdUI* pCmdUI);
	afx_msg void OnOptReplace();
	afx_msg void OnFileOpenspecial();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
private:
	BOOL m_bFirstInit;
	// Variables
	BOOL			m_bItemsChanged;
	BOOL			m_bItemsAdded;
	BOOL			m_bErrorsLogged;
	BOOL			m_bIEPresent;
	BOOL			m_bInternalLoadFail;
	BOOL			m_bGotSomeData;
	BOOL			m_bLoadFailed;
	CStdioFile*		m_pFile;
	CStdioFile*		m_pLogFile;
	CStatDlg*		m_pStatDlg;
	CString			m_strIEPath;
	CString			m_strErrLog;
	// Functions
	void Detect_Browsers(void);
	void Report_Fatal_Error(LPCTSTR text);
	void Report_HTML_Err(CString& html, LPCTSTR err);
	void Clean_Folder(LPCTSTR pszPath);
	void Do_Import();
	void Export_Text(void);
	void End_Import_Log();
	void Do_Import_Cleanup();
	void Query_Show_Log(BOOL bFatal);
	BOOL Create_URL_File(LPCTSTR szPath, UINT nItemOffset);
	BOOL Validate_Path(LPCTSTR pszPath);
	BOOL Start_Import_Log(void);
	BOOL Resolve_Dupes(void);
	BOOL Export_MSIE_Path(UINT nPathOffset);
	BOOL Parse_NS_Folder(LPCTSTR pszOldPath = NULL, LPCTSTR pszNewPath = NULL);
	BOOL Scan_Folder(LPCTSTR lpszPath = NULL);
	UINT Parse_NS_File(LPCTSTR lpszFileName);
	// Global callback functions used by qsort
	friend int __cdecl Compare_Items(const void* p1, const void* p2);
	friend int __cdecl Compare_Dupes(const void* p1, const void* p2);
	friend class CWeedDlg;
};


/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_SUDOC_H__A03AD40C_1CC5_11D1_84C6_444553540000__INCLUDED_)
