// SUView.cpp : implementation of the CSUView class
//
//
// SyncURLs Version 1.4
// Copyright  1998 Ziff-Davis Inc.
// First published in PC Magazine, US Edition
// January 6, 1998
// Written by Gregory A. Wolking
//

#include "stdafx.h"
#include "SyncURLs.h"
#include <afxpriv.h>

#include "MyTree.h"
#include "MyList.h"
#include "PropDlg.h"
#include "WeedDlg.h"
#include "StatDlg.h"
#include "FindDlg.h"
#include "ScrllDlg.h"
#include "RenDlg.h"
#include "SUDoc.h"
#include "SUView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// CSUView

IMPLEMENT_DYNCREATE(CSUView, CFormView)

BEGIN_MESSAGE_MAP(CSUView, CFormView)
	//{{AFX_MSG_MAP(CSUView)
	ON_WM_CREATE()
	ON_COMMAND(ID_EDIT_RENAME, OnEditRename)
	ON_UPDATE_COMMAND_UI(ID_EDIT_RENAME, OnUpdateEditRename)
	ON_NOTIFY(TVN_SELCHANGED, IDC_tvnFolders, OnSelchangedtvnFolders)
	ON_NOTIFY(NM_RCLICK, IDC_tvnFolders, OnRclicktvnFolders)
	ON_NOTIFY(TVN_BEGINDRAG, IDC_tvnFolders, OnBegindragtvnFolders)
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_COMMAND(ID_EDIT_DEL, OnEditDel)
	ON_UPDATE_COMMAND_UI(ID_EDIT_DEL, OnUpdateEditDel)
	ON_COMMAND(ID_OPT_CONFIRM_BOOKMARK, OnOptConfirmBookmark)
	ON_UPDATE_COMMAND_UI(ID_OPT_CONFIRM_BOOKMARK, OnUpdateOptConfirmBookmark)
	ON_NOTIFY(NM_CLICK, IDC_tvnFolders, OnClicktvnFolders)
	ON_COMMAND(ID_OPT_CONFIRM_FOLDER, OnOptConfirmFolder)
	ON_UPDATE_COMMAND_UI(ID_OPT_CONFIRM_FOLDER, OnUpdateOptConfirmFolder)
	ON_COMMAND(ID_EDIT_NEW, OnEditNew)
	ON_COMMAND(ID_VIEW_PROPERTIES, OnViewProperties)
	ON_UPDATE_COMMAND_UI(ID_VIEW_PROPERTIES, OnUpdateViewProperties)
	ON_COMMAND(ID_EDIT_WEED, OnEditWeed)
	ON_UPDATE_COMMAND_UI(ID_EDIT_WEED, OnUpdateEditWeed)
	ON_WM_RBUTTONDOWN()
	ON_COMMAND(ID_VIEW_EXPAND, OnViewExpand)
	ON_COMMAND(ID_VIEW_COLLAPSE, OnViewCollapse)
	ON_COMMAND(ID_OPT_AUTO, OnOptAuto)
	ON_COMMAND(ID_EDIT_FIND, OnEditFind)
	ON_COMMAND(ID_EDIT_FIND_NEXT, OnEditFindNext)
	ON_UPDATE_COMMAND_UI(ID_EDIT_FIND_NEXT, OnUpdateEditFindNext)
	ON_UPDATE_COMMAND_UI(ID_EDIT_FIND, OnUpdateEditFind)
	ON_COMMAND(ID_EXPORT_MANUAL, OnExportManual)
	ON_COMMAND(ID_EXPORT_NS, OnExportNs)
	ON_COMMAND(ID_EXPORT_MSIE, OnExportMsie)
	ON_COMMAND(ID_EXPORT_BOTH, OnExportBoth)
	ON_COMMAND(ID_EXPORT_NSC, OnExportNsc)
	ON_COMMAND(ID_OPT_FOLLOW, OnOptFollow)
	ON_UPDATE_COMMAND_UI(ID_OPT_FOLLOW, OnUpdateOptFollow)
	ON_NOTIFY(TVN_KEYDOWN, IDC_tvnFolders, OnKeydowntvnFolders)
	ON_NOTIFY(NM_DBLCLK, IDC_tvnFolders, OnDblclktvnFolders)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSUView construction/destruction

CSUView::CSUView()
	: CFormView(CSUView::IDD)
{
	//{{AFX_DATA_INIT(CSUView)
	//}}AFX_DATA_INIT
	// TODO: add construction code here
	m_intSearchIn = 0;
	m_strSearchText.Empty();
	m_bSearchCase = FALSE;
	m_bSearchStart = FALSE;
	m_pFile = NULL;
}

CSUView::~CSUView()
{
	CWinApp* pMyApp = AfxGetApp();
	pMyApp->WriteProfileInt("Settings", "ConfirmBookmarkDel", m_bConfirmBookmarkDelete);
	pMyApp->WriteProfileInt("Settings", "ConfirmFolderDel", m_bConfirmFolderDelete);
	pMyApp->WriteProfileInt("Settings", "Follow", m_bFollow);
}

void CSUView::DoDataExchange(CDataExchange* pDX)
{
	CFormView::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CSUView)
	DDX_Control(pDX, IDC_tvnFolders, m_tvnFolders);
	//}}AFX_DATA_MAP
}

BOOL CSUView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CFormView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CSUView diagnostics

#ifdef _DEBUG
void CSUView::AssertValid() const
{
	CFormView::AssertValid();
}

void CSUView::Dump(CDumpContext& dc) const
{
	CFormView::Dump(dc);
}

CSUDoc* CSUView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CSUDoc)));
	return (CSUDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CSUView message handlers

/////////////////////////
// CFormView handler overrides

void CSUView::OnInitialUpdate() 
{
	CFormView::OnInitialUpdate();
	
	GetParentFrame()->RecalcLayout();
	SetScaleToFitSize(CSize(1,1));
	if (m_imgTree.m_hImageList == NULL)							// If image list doesn't exist yet,
	{
		m_imgTree.Create(IDB_BITMAP2, 16, 0, 0x00FFFFFF);		// Create it.
		m_tvnFolders.SetImageList(&m_imgTree, TVSIL_NORMAL);	// Attach it to the tree control.
		m_imgState.Create(IDB_BITMAP1, 16, 0, 0x00FFFFFF);
		m_tvnFolders.SetImageList(&m_imgState, TVSIL_STATE);
	}
}

void CSUView::OnDraw(CDC* pDC) 
{
	if (m_tvnFolders.m_hWnd == NULL)
		return;
	CRect wnd_rect;

	GetWindowRect(wnd_rect);				// Get window coordinates.
	ScreenToClient(wnd_rect);				// Convert to client coordinates.
	m_tvnFolders.MoveWindow(wnd_rect);		// Adjust tree control to fill the window.
}

void CSUView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	BeginWaitCursor();								// Let user know we're working.
	m_pMyDoc = (CSUDoc*) GetDocument();				// Make sure pointer to current document is valid.
	m_tvnFolders.DeleteAllItems();					// Zap the tree control.
	m_pMyDoc->Reset_Path_Flags();					// Prepare for recursion.
	Process_Path(0, TVI_ROOT);						// Fill the tree control.
	m_hCurItem = m_tvnFolders.GetRootItem();		// Set current item handle to root folder
	m_tvnFolders.Expand(m_hCurItem, TVE_EXPAND);	// Expand the root folder.
	// If hint parameter is non-zero, update was due to a folder move,
	if (lHint)
		// so make sure the newly-moved folder is visible.
		Locate_Folder(m_hCurItem, lHint);
	EndWaitCursor();								// Restore normal cursor.
}

int CSUView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CFormView::OnCreate(lpCreateStruct) == -1)
		return -1;

	CWinApp* pMyApp = AfxGetApp();
	m_hCurItem = NULL;
	m_hDragSource = NULL;
	m_hDragDest = NULL;
	m_pMyDoc = NULL;
	m_bIsDragging = FALSE;
	// Retrieve settings from Registry.
	m_bConfirmBookmarkDelete = pMyApp->GetProfileInt("Settings", "ConfirmBookmarkDel", TRUE);
	m_bConfirmFolderDelete = pMyApp->GetProfileInt("Settings", "ConfirmFolderDel", TRUE);
	m_bFollow = pMyApp->GetProfileInt("Settings", "Follow", FALSE);
	// Clean up any invalid numeric Registry data.
	m_intScrollRate = pMyApp->GetProfileInt("Settings", "ScrollRate", 0);
	if (m_intScrollRate > 100)
	{
		m_intScrollRate = 100;
		pMyApp->WriteProfileInt("Settings", "ScrollRate", 100);
	}
	if (m_intScrollRate < 0)
	{
		m_intScrollRate = 0;
		pMyApp->WriteProfileInt("Settings", "ScrollRate", 0);
	}
	return 0;
}

				///////////////////////////////////
				// Menu command and UI handlers. //
				///////////////////////////////////

///////////////////////////
// "Edit" menu commands. //
///////////////////////////


// "Rename"
void CSUView::OnEditRename() 
{
	char		szBuff[_MAX_PATH];
	BOOL		bIsBookmark,
				bDupe;
	UINT		nItemIndex;
	CString		strOldName,
				strNewName,
				strWork;
	TV_ITEM		tvItem;
	HTREEITEM	hCurItem,
				hParentItem,
				hWorkItem;

	hCurItem = m_tvnFolders.GetSelectedItem();
	if (hCurItem == NULL)
		return;
	hParentItem = m_tvnFolders.GetParentItem(hCurItem);
	tvItem.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_PARAM;
	tvItem.hItem = hCurItem;
	tvItem.cchTextMax = sizeof(szBuff);
	tvItem.pszText = szBuff;
	m_tvnFolders.GetItem(&tvItem);
	bIsBookmark = (tvItem.iImage == 1);
	nItemIndex = tvItem.lParam;
	strOldName = szBuff;
	CRenDlg dlg(this);
	dlg.m_txtNewName = strOldName;
	if (bIsBookmark)
		dlg.m_strType = "Bookmark";
	else
		dlg.m_strType = "Folder";
	// Loop repeats until either a valid name is given or the user cancels.
	while (TRUE)
	{
		if (dlg.DoModal() != IDOK)
			return;
		strNewName = dlg.m_txtNewName;
		strNewName.TrimRight();
		strNewName.TrimLeft();
		if (strNewName.IsEmpty())	// Error if new name is blank.
		{
			AfxMessageBox(IDP_NO_BLANK_NAME, MB_ICONINFORMATION);
			dlg.m_txtNewName = strOldName;
			continue;
		}
		// Error if name begings with a "."
		if (strNewName.Left(1) == '.')
		{
			AfxMessageBox(IDP_NO_PERIOD, MB_ICONINFORMATION);
			continue;
		}
		// Clean up illegal characters.
		m_pMyDoc->Fix_FileName(strNewName);
		// If names are identical after all of this, there's nothing to do!
		if (strOldName == strNewName)
			return;
		// For folders, make sure that name doesn't have the ".URL" extension.
		if (!bIsBookmark)
		{
			strWork = strNewName.Right(4);
			if (strWork.CompareNoCase(".url") == 0)
			{
				AfxMessageBox(IDP_NO_URL_EXT, MB_ICONINFORMATION);
				dlg.m_txtNewName = strNewName;
				continue;
			}
		}
		// Check for existing item of the same type with the same name.
		hWorkItem = m_tvnFolders.GetChildItem(hParentItem);
		bDupe = FALSE;
		do
		{
			if (hWorkItem != hCurItem)
			{
				tvItem.hItem = hWorkItem;
				m_tvnFolders.GetItem(&tvItem);
				if (bIsBookmark && tvItem.iImage != 1)
					continue;
				if (!bIsBookmark && tvItem.iImage == 1)
					continue;
				if (strNewName.CompareNoCase(szBuff) == 0)
				{
					bDupe = TRUE;
					break;
				}
			}
		} while ((hWorkItem = m_tvnFolders.GetNextSiblingItem(hWorkItem)) != NULL);
		if (bDupe)	// Error and repeat if dupe item found.
		{
			AfxMessageBox(IDP_UNIQUE, MB_ICONINFORMATION);
			continue;
		}
		// If we got this far, we're ready to rename.
		if (bIsBookmark)	// Handle bookmark.
		{
			*m_pMyDoc->m_pDocElements[nItemIndex].pPage = strNewName;	// Update the document.
			m_pMyDoc->SetModifiedFlag();
			m_tvnFolders.SetItemText(hCurItem, strNewName);		// Update the item's text.
			m_hDragSource = hCurItem;							// Move the item to its own
			m_hDragDest = hParentItem;							// Parent folder. This puts the
			Move_Bookmark(TRUE);								// renamed item back in sorted order.
			break;	// All done, so exit the loop.
		}
		// For folders, make sure we can update it. (If resulting path
		// would be too long, Rename_Folder() returns false).
		if (!m_pMyDoc->Rename_Folder(nItemIndex, strNewName))
		{
			AfxMessageBox(IDP_PATH_TOO_LONG, MB_ICONINFORMATION);
			continue;
		}
		// Update folder item's text.
		m_tvnFolders.SetItemText(hCurItem, strNewName);
		// Move it to its own parent so it goes into proper sorted order.
		m_hDragSource = hCurItem;
		m_hDragDest = hParentItem;
		Move_Folder(TRUE);
		break;
	}
}

void CSUView::OnUpdateEditRename(CCmdUI* pCmdUI) 
{
	BOOL		bStatus = FALSE;
	TV_ITEM		tvItem;
	CString		strMessage;
	
	if (m_hCurItem)
	{
		tvItem.hItem = m_hCurItem;
		tvItem.mask = TVIF_IMAGE;
		m_tvnFolders.GetItem(&tvItem);
		switch(tvItem.iImage)
		{
		case 0:
			strMessage = "&Rename folder\tCtrl+R";
			bStatus = TRUE;
			break;
		case 2:
			strMessage = "&Rename folder\tCtrl+R";
			break;
		case 1:
			strMessage = "&Rename bookmark\tCtrl+R";
			bStatus = TRUE;
			break;
		case 3:
			strMessage = "&Rename bookmark\tCtrl+R";
		}
	}
	pCmdUI->SetText(strMessage);
	pCmdUI->Enable(bStatus);
}

// "Delete" command.
void CSUView::OnEditDel() 
{
	BOOL		bAll;
	UINT		nOffset,
				nOldState;
	TV_ITEM		tvItem;
	HTREEITEM	hItem,
				hParent;

	hItem = m_tvnFolders.GetSelectedItem();				// Get handle to current item.
	hParent = m_tvnFolders.GetParentItem(hItem);
	if (IsBookmark(hItem))
	{
		if (m_bConfirmBookmarkDelete)					// Give confirmation prompt if appropriate.
		{
			if (AfxMessageBox(IDP_CONFIRM_BOOKMARK_DEL, MB_YESNO | MB_ICONQUESTION) != IDYES)
				return;
		}
		BeginWaitCursor();									// Put up the busy cursor
		m_nDeletedItem = m_tvnFolders.GetItemData(hItem);	// Get hItem's nOffset into the document.
		m_pMyDoc->Delete_Item(m_nDeletedItem);				// Remove it from the document.
		m_tvnFolders.DeleteItem(hItem);						// Remove the hItem from the tree.
		Adjust_Bookmarks(m_tvnFolders.GetRootItem());		// Adjust offsets for other items.
		EndWaitCursor();									// Restore normal cursor.
	}
	else
	{
		if (m_bConfirmFolderDelete)
		{
			if (AfxMessageBox(IDP_CONFIRM_FOLDER_DEL, MB_YESNO | MB_ICONQUESTION) != IDYES)
				return;
		}
		BeginWaitCursor();
		nOffset = m_tvnFolders.GetItemData(hItem);
		m_pMyDoc->Delete_Folder(nOffset);
		m_tvnFolders.DeleteItem(hItem);
		Adjust_Folder_Offsets(m_tvnFolders.GetRootItem(), nOffset);
		tvItem.mask = TVIF_CHILDREN | TVIF_STATE | TVIF_PARAM;
		tvItem.hItem = hParent;
		m_tvnFolders.GetItem(&tvItem);
		nOldState = (tvItem.state & TVIS_STATEIMAGEMASK) >> 12;
		nOffset = (UINT) tvItem.lParam;
		if (nOldState == 3)					// If parent folder was partially selected,
		{
			if (tvItem.cChildren == 0)		// and it has no remaining children,
				nOldState = 2;				// it becomes fully selected.
			else							// Otherwise, check status of remaining children.
			{
				bAll = TRUE;					// Default = All children selected.
				hItem = m_tvnFolders.GetChildItem(hParent);	// Get handle of first child item.
				tvItem.mask = TVIF_STATE | TVIF_IMAGE;		// Mask for info required for each item.
				do
				{
					tvItem.hItem = hItem;					// Set handle.
					m_tvnFolders.GetItem(&tvItem);			// Retrieve info.
					if (tvItem.iImage != 1)					// Is item a folder? 
					{
						if ((tvItem.state & TVIS_STATEIMAGEMASK) != 0x2000)	// Is it not fully selected?
						{
							bAll = FALSE;					// Then not all children are selected.
							break;							// Done looking.
						}
					}
					hItem = m_tvnFolders.GetNextSiblingItem(hItem);	// Next item.
				} while (hItem != NULL);	// Loop until no items left.
				if (bAll)				// If all children were selected,
					nOldState = 2;		// parent becomes fully selected.
			}
			if (nOldState != 3)			// If selection state changed,
			{
				m_pMyDoc->m_pPathList[nOffset].nSelected = nOldState;	// Update the document.
				tvItem.mask = TVIF_STATE;								// Update the item.
				tvItem.state = nOldState << 12;
				tvItem.stateMask = TVIS_STATEIMAGEMASK;
				tvItem.hItem = hParent;
				m_tvnFolders.SetItem(&tvItem);
				Set_State_Up(hParent);								// Update the rest of the tree.
			}
		}
		EndWaitCursor();
	}
}

void CSUView::OnUpdateEditDel(CCmdUI* pCmdUI) 
{
	BOOL		bStatus = FALSE;
	TV_ITEM		tvItem;
	HTREEITEM	hItem;

	if (m_pMyDoc->m_nItemCount)
	{
		hItem = m_tvnFolders.GetSelectedItem();
		if (hItem != NULL)
		{
			tvItem.mask = TVIF_IMAGE | TVIF_CHILDREN;
			tvItem.hItem = hItem;
			m_tvnFolders.GetItem(&tvItem);
			bStatus = ((tvItem.iImage == 1) || (tvItem.iImage == 0 && tvItem.cChildren == 0));
			if (tvItem.iImage == 1)
				pCmdUI->SetText("&Delete bookmark\tDel");
			else
				pCmdUI->SetText("&Delete folder\tDel");
		}
	}
	pCmdUI->Enable(bStatus);
}

// "New folder" command.
void CSUView::OnEditNew() 
{
	char				buff[_MAX_PATH];
	BOOL				bValid,
						bModified;
	UINT				nOffset,
						nSuffix = 0;
	CString				strNewPath,
						strNewName;
	TV_ITEM				tvItem1,
						tvItem2;
	HTREEITEM			hParent,
						hCurItem;
	TV_INSERTSTRUCT		tvInsert;

	hCurItem = m_tvnFolders.GetSelectedItem();	// Get handle to current item.
	if (hCurItem == NULL)						// Done if no item selected.
		return;
	tvItem1.mask = TVIF_IMAGE;					// Get item info.
	tvItem1.hItem = hCurItem;
	m_tvnFolders.GetItem(&tvItem1);
	if (tvItem1.iImage == 1)					// If it's a bookmark,
		hParent = m_tvnFolders.GetParentItem(hCurItem);		// get its parent folder.
	else
		hParent = hCurItem;
	tvItem1.mask = TVIF_CHILDREN | TVIF_PARAM;	// Retrieve item info.
	tvItem1.hItem = hParent;
	m_tvnFolders.GetItem(&tvItem1);
	nOffset = (UINT) tvItem1.lParam;			// Get offset to parent folder's path.
	if (nOffset != 0)							// If it's not the root,
		strNewPath = *m_pMyDoc->m_pPathList[nOffset].pPath;	// Copy it.
	else										// Otherwise,
		strNewPath.Empty();						// parent path is empty.
	bModified = m_pMyDoc->IsModified();
	if ((m_pMyDoc->m_iBasePathLen + strNewPath.GetLength()) > 180)	// Check the length.
	{
		AfxMessageBox(IDP_PATH_TOO_LONG_CREATE, MB_ICONINFORMATION);
		return;
	}
	// Determine and assign a default name.
	do
	{
		// Set proposed default name based on the numeric suffix.
		if (nSuffix)
			strNewName.Format("New Folder (%u)", nSuffix);
		else
			strNewName = "New Folder";
		// Done if folder has no children.
		if (tvItem1.cChildren == 0)
			break;
		// Start search with first child item.
		hCurItem = m_tvnFolders.GetChildItem(hParent);
		bValid = TRUE;	// Default status = name is valid.
		tvItem2.mask = TVIF_IMAGE | TVIF_TEXT;
		tvItem2.pszText = buff;
		tvItem2.cchTextMax = sizeof(buff);
		while (hCurItem != NULL)
		{
			tvItem2.hItem = hCurItem;			// Retrieve info for current item.
			m_tvnFolders.GetItem(&tvItem2);
			if (tvItem2.iImage != 1)			// Skip if it's not a folder.
			{
				if (strNewName.CompareNoCase(tvItem2.pszText) == 0)	// See if name matches.
				{
					bValid = FALSE;				// If it does, set invalid status and and stop looking.
					break;
				}
			}
			hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);	// Continue with next item.
		}
		++nSuffix;	// Next numeric suffix.
	} while (!bValid);	// Loop until valid name is found.
	strNewPath += strNewName;	// Create new name.
	strNewPath += '\\';
	// Add new path to the document; report error if necessary.
	if (!m_pMyDoc->Add_Path(strNewPath, FALSE, &nOffset))
	{
		AfxMessageBox(IDP_PATH_CREATE_FAIL, MB_ICONINFORMATION);
		return;
	}
	// Now create the new folder item in the tree.
	tvInsert.hParent = hParent;						// Set handle to parent item.
	lstrcpyn(buff, strNewName, sizeof(buff));		// Copy new name to buffer.
	tvInsert.hInsertAfter = TVI_SORT;				// New folder inserted in sorted order.
	tvInsert.item.mask = TVIF_IMAGE | TVIF_STATE | TVIF_TEXT | TVIF_PARAM;
	tvInsert.item.iImage = 0;						// Item has a folder icon.
	tvInsert.item.cchTextMax = strNewName.GetLength() + 1;	// Set name length.
	tvInsert.item.pszText = buff;					// Set pointer to name.
	tvInsert.item.state = (2 << 12);				// New folder is selected by default.
	tvInsert.item.stateMask = TVIS_STATEIMAGEMASK;
	tvInsert.item.lParam = (LPARAM) nOffset;		// Set offset to folder's path in the document.
	hCurItem = m_tvnFolders.InsertItem(&tvInsert);	// Insert the item into the tree.
	if (hCurItem == NULL)							// If insert failed,
	{
		m_pMyDoc->Delete_Folder(nOffset);			// remove new path from the document and give error message.
		if (!bModified)								// If document wasn't dirty before we started,
			m_pMyDoc->SetModifiedFlag(FALSE);		// mark it as "clean" now.
		AfxMessageBox(IDP_TREE_CREATE_FAIL, MB_ICONINFORMATION);
	}
	else											// Otherwise,
	{
		m_tvnFolders.EnsureVisible(hCurItem);		// make the new item visible,
		Set_State_Up(hCurItem);						// adjust selection state of parent folders,
		m_tvnFolders.SelectItem(hCurItem);			// select the new folder,
		m_pMyDoc->SetModifiedFlag(TRUE);			// and mark document as dirty.
	}
}

// "Weed Duplicates" command
void CSUView::OnEditWeed() 
{
	if (Weed_Duplicates())
		OnUpdate(NULL, 0, NULL);
}

void CSUView::OnUpdateEditWeed(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pMyDoc->m_nItemCount >= 2);
}

void CSUView::OnUpdateEditFindNext(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(!m_strSearchText.IsEmpty() && m_pMyDoc->m_nItemCount > 0);	
}

void CSUView::OnUpdateEditFind(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pMyDoc->m_nItemCount > 0);	
}

// "Find" command
void CSUView::OnEditFind() 
{
	CFindDlg	dlg;							// Create search criteria dialog.
	HTREEITEM	hItem;

	dlg.m_bIgnoreCase = m_bSearchCase;		// Load it with current criteria.
	dlg.m_intLook = m_intSearchIn;
	dlg.m_strTarget = m_strSearchText;
	dlg.m_intSearchStart = m_bSearchStart;
	if (dlg.DoModal() != IDOK)				// Launch the dialog.
		return;								// Done if user cancelled.
	m_bSearchCase = dlg.m_bIgnoreCase;		// Store new criteria.
	m_intSearchIn = dlg.m_intLook;
	m_strSearchText = dlg.m_strTarget;
	m_bSearchStart = dlg.m_intSearchStart;
	if (m_bSearchStart)
		hItem = m_tvnFolders.GetSelectedItem();
	else
		hItem = m_tvnFolders.GetRootItem();
	if (!Search_Tree(hItem))				// Do the search.
		AfxMessageBox(IDP_NOT_FOUND, MB_ICONINFORMATION);	// Error message if appropriate.
}

// "Find Next" command.
void CSUView::OnEditFindNext() 
{
	HTREEITEM hItem;

	hItem = m_tvnFolders.GetSelectedItem();
	hItem = Next_Tree_Item(hItem);
	if (!Search_Tree(hItem))					// Continue existing search.
		AfxMessageBox(IDP_NOT_FOUND, MB_ICONINFORMATION);	// Error message if appropriate.
}

///////////////////////////
// Export Menu commands. //
///////////////////////////

// "Other BOOKMARK.HTM..." command.
void CSUView::OnExportManual() 
{
	CString		strOldName,
				strOldTitle;

	if (!m_pMyDoc->Do_Export_Warning())	// Give user a chance to bail if no folders are selected.
		return;
	// Create a Save As file dialog with appropriate title and filter settings.
	CFileDialog dlg(FALSE, ".htm", "Bookmark.htm", OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"HTML Files (*.htm; *.html)|*.htm; *.html||", NULL);
	dlg.m_ofn.lpstrTitle = "Export HTML File";
	// Launch the File dialog.
	if (dlg.DoModal() != IDOK)
		return;		// Done if user didn't say OK.
	// Save existing title and filename strings.
	strOldName = m_pMyDoc->m_strNSPath;
	strOldTitle = m_pMyDoc->m_strNSPageTitle;
	// Install new title and filename strings.
	m_pMyDoc->m_strNSPath = dlg.GetPathName();
	m_pMyDoc->m_strNSPageTitle = "Bookmarks exported by SyncURLs";
	// Export to the specified file.
	Write_NS_File(FALSE);
	// Restore original title and filename strings.
	m_pMyDoc->m_strNSPath = strOldName;
	m_pMyDoc->m_strNSPageTitle = strOldTitle;
}

// "Netscape 3.x" command.
void CSUView::OnExportNs() 
{
	if (!m_pMyDoc->Do_Export_Warning())	// Give user a chance to bail if no folders are selected.
		return;
	Write_NS_File();	
}

// "Netscape 4.x" command.
void CSUView::OnExportNsc() 
{
	if (!m_pMyDoc->Do_Export_Warning())	// Give user a chance to bail if no folders are selected.
		return;
	CString strOldPath(m_pMyDoc->m_strNSPath);
	m_pMyDoc->m_strNSPath = m_pMyDoc->m_strNSCPath;
	Write_NS_File();
	m_pMyDoc->m_strNSPath = strOldPath;
}

// "All Browsers" command.
void CSUView::OnExportBoth() 
{
	if (!m_pMyDoc->Do_Export_Warning())		// Warn if no folders selected.
		return;
	if (m_pMyDoc->m_bNSPresent)				// NS 3.x detected?
		Write_NS_File();					// Export if so.
	if (m_pMyDoc->m_bNSCPresent)			// NS 4.x detected?
	{
		CString strOldPath(m_pMyDoc->m_strNSPath);			// Save NS 3.x path.
		m_pMyDoc->m_strNSPath = m_pMyDoc->m_strNSCPath;		// Install NS 4.x path.
		Write_NS_File();									// Export.
		m_pMyDoc->m_strNSPath = strOldPath;					// Restore NS 3.x path.
	}
	m_pMyDoc->Do_MSIE_Export();				// Export to MSIE, if detected.
}

// "MSIE" command.
void CSUView::OnExportMsie() 
{
	if (!m_pMyDoc->Do_Export_Warning())
		return;
	m_pMyDoc->Do_MSIE_Export();
}

//////////////////////////
// View menu commmands. //
//////////////////////////

void CSUView::OnViewExpand() 
{
	Expand_All(NULL, TRUE);
}

void CSUView::OnViewCollapse() 
{
	Expand_All(NULL, FALSE);
}

// "Properties" command.
// NOTE: UI enables this command only if a bookmark is selected in the tree.
void CSUView::OnViewProperties() 
{
	UINT			nOffset;
	CPropDlg		dlg;
	HTREEITEM		hItem;
	DOC_ELEMENT		*pDocItem;

	hItem = m_tvnFolders.GetSelectedItem();							// Get index of current item.
	if (hItem == NULL)												// Nothing to do if no item selected.
		return;
	nOffset = m_tvnFolders.GetItemData(hItem);						// Get array offset of item.
	pDocItem = &m_pMyDoc->m_pDocElements[nOffset];					// Get pointer to document item.
	dlg.m_strPageName = *pDocItem->pPage;							// Get page name.
	dlg.m_strURL = *pDocItem->pUrl;									// Get URL.
	Format_Date_Field(pDocItem->nAdd, dlg.m_strAddDate);		// Retrieve and format date fields.
	Format_Date_Field(pDocItem->nMod, dlg.m_strModDate);
	Format_Date_Field(pDocItem->nVisit, dlg.m_strVisitDate);
	if (dlg.DoModal() == IDOK)					// Launch dialog. If user said "OK",
	{
		if (*pDocItem->pUrl != dlg.m_strURL)		// see if URL string changed. If it did,
		{
			*pDocItem->pUrl = dlg.m_strURL;		// install the updated string and
			m_pMyDoc->SetModifiedFlag();		// mark the document as "dirty'.
		}
	}
}

void CSUView::OnUpdateViewProperties(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(IsBookmark(m_hCurItem));
}

////////////////////////////
// Options menu commands. //
////////////////////////////

void CSUView::OnOptConfirmBookmark() 
{
	m_bConfirmBookmarkDelete = !m_bConfirmBookmarkDelete;	
}

void CSUView::OnUpdateOptConfirmBookmark(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_bConfirmBookmarkDelete);	
}

void CSUView::OnOptConfirmFolder() 
{
	m_bConfirmFolderDelete = !m_bConfirmFolderDelete;	
}

void CSUView::OnUpdateOptConfirmFolder(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_bConfirmFolderDelete);	
}

void CSUView::OnOptAuto() 
{
	CScrllDlg dlg;

	dlg.m_intScrollRate = 100 - m_intScrollRate;
	if (dlg.DoModal() != IDOK)
		return;
	m_intScrollRate = 100 - dlg.m_intScrollRate;
	AfxGetApp()->WriteProfileInt("Settings", "ScrollRate", m_intScrollRate);
}

void CSUView::OnOptFollow() 
{
	m_bFollow = !m_bFollow;
}

void CSUView::OnUpdateOptFollow(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_bFollow);	
}


				////////////////////////////////////
				// Control notification handlers. //
				////////////////////////////////////

// Tree control selection changed.
void CSUView::OnSelchangedtvnFolders(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	
	m_hCurItem = pNMTreeView->itemNew.hItem;
	*pResult = 0;
}

// Provides right-click pop-up menu for the tree control.
void CSUView::OnRclicktvnFolders(NMHDR* pNMHDR, LRESULT* pResult) 
{
	UINT		nFlags;
	CPoint		point;
	TV_ITEM		tvItem;
	HTREEITEM	hItem;

	GetCursorPos(&point);
	ScreenToClient(&point);
	hItem = m_tvnFolders.HitTest(point, &nFlags);
	if (hItem != NULL)
	{
		tvItem.mask = TVIF_IMAGE | TVIF_CHILDREN;
		tvItem.hItem = hItem;
		m_tvnFolders.GetItem(&tvItem);
		ClientToScreen(&point);				// Get client coordinates for menu position.
		m_tvnFolders.SelectItem(hItem);		// Make sure item is selected.
		CMenu menu;							// Create the popup menu.
		menu.CreatePopupMenu();
		if (tvItem.iImage == 1)			// Build menu based on item type.
		{
			menu.AppendMenu(MF_STRING, ID_EDIT_RENAME, "&Rename Bookmark");
			menu.AppendMenu(MF_STRING, ID_EDIT_DEL, "&Delete Bookmark");
			menu.AppendMenu(MF_SEPARATOR);
		}
		else
		{
			if (tvItem.iImage != 2)
			{
				menu.AppendMenu(MF_STRING, ID_EDIT_RENAME, "&Rename Folder");
				if (tvItem.iImage == 0 && tvItem.cChildren == 0)
					menu.AppendMenu(MF_STRING, ID_EDIT_DEL, "&Delete Folder");
				menu.AppendMenu(MF_SEPARATOR);
			}
		}
		menu.AppendMenu(MF_STRING, ID_EDIT_NEW, "&New folder");
		if (tvItem.iImage == 1)
		{
			menu.AppendMenu(MF_SEPARATOR);
			menu.AppendMenu(MF_STRING, ID_VIEW_PROPERTIES, "&Properties");
		}
		// Launch the menu.
		menu.TrackPopupMenu(TPM_LEFTALIGN | TPM_LEFTBUTTON, point.x, point.y, GetParentFrame(), NULL);
	}
	*pResult = 0;
}

// Initiates bookmark/folder drag.
void CSUView::OnBegindragtvnFolders(NMHDR* pNMHDR, LRESULT* pResult) 
{
	CPoint			point;
	HTREEITEM		hDragItem;
	NM_TREEVIEW		*pNMTreeView = (NM_TREEVIEW*)pNMHDR;

	point = pNMTreeView->ptDrag;
	hDragItem = pNMTreeView->itemNew.hItem;
	if (!IsLocked(hDragItem))
	{
		m_bIsDragging = TRUE;
		m_hDragSource = hDragItem;
		m_hDragDest = NULL;
		m_tvnFolders.SelectItem(hDragItem);
		m_pDragImage = m_tvnFolders.CreateDragImage(hDragItem);
		m_pDragImage->DragShowNolock(TRUE);
		m_pDragImage->SetDragCursorImage(0, CPoint(0,0));
		m_pDragImage->BeginDrag(0, CPoint(0,0));
		m_pDragImage->DragMove(point);
		m_pDragImage->DragEnter(this, point);
		SetCapture();
	}
	else
	{
		AfxMessageBox(IDP_FOLDER_LOCKED, MB_ICONINFORMATION);
		m_tvnFolders.SelectItem(hDragItem);
		m_hCurItem = hDragItem;
	}
	*pResult = 0;
}

// Handles mouse click in the tree control.
void CSUView::OnClicktvnFolders(NMHDR* pNMHDR, LRESULT* pResult) 
{
	UINT		nFlags,
				nState;
	CPoint		point;
	TV_ITEM		tvItem;
	HTREEITEM	hItem;

	GetCursorPos(&point);							// Get mouse position.
	ScreenToClient(&point);							// Convert to client coordinates.
	hItem = m_tvnFolders.HitTest(point, &nFlags);	// Ask the tree control what was clicked on.
	if (hItem != NULL)								// Nothing to do if not an hItem.
	{
		if (nFlags & TVHT_ONITEMSTATEICON)		// If it was the state image (selection box)
		{
			tvItem.mask = TVIF_STATE;				// Get item state.
			tvItem.hItem = hItem;
			m_tvnFolders.GetItem(&tvItem);
			nState = tvItem.state >> 12;			// Adjust value.
			if (nState > 1)						// If already checked,
				nState = 1;						// clear it.
			else								// Otherwise,
				nState = 2;						// check it.
			Set_State_Down(hItem, nState > 1);	// Set selection state for children.
			Set_State_Up(hItem);				// Adjust selection state for parents.
		}
	}
	*pResult = 0;
}

// Handle double-click in the tree control.
void CSUView::OnDblclktvnFolders(NMHDR* pNMHDR, LRESULT* pResult) 
{
	UINT		nFlags;
	CPoint		point;
	TV_ITEM		tvItem;
	HTREEITEM	hItem;
	
	GetCursorPos(&point);							// Get mouse position.
	ScreenToClient(&point);							// Convert to client coordinates.
	hItem = m_tvnFolders.HitTest(point, &nFlags);	// Ask the tree control what was clicked on.
	if (hItem != NULL)								// Nothing to do if in the open area of the control.
	{
		if (nFlags & TVHT_ONITEM)					// If it was the item icon or its label,
		{
			tvItem.mask = TVIF_IMAGE;				// Get item's image.
			tvItem.hItem = hItem;
			m_tvnFolders.GetItem(&tvItem);
			if (tvItem.iImage == 1)					// If it's a bookmark,
				OnViewProperties();					// launch the Properties dialog.
		}
	}
	*pResult = 0;
}

// Handles key press while the Tree Control has the focus.
// This enables use of the spacebar to select/deselect folders.
void CSUView::OnKeydowntvnFolders(NMHDR* pNMHDR, LRESULT* pResult) 
{
	UINT		nState;
	TV_ITEM		tvItem;
	HTREEITEM	hItem;
	NMTVKEYDOWN	*pTVKeyDown = (NMTVKEYDOWN*)pNMHDR;

	if (pTVKeyDown->wVKey == VK_SPACE)					// Was it the spacebar?
	{
		hItem = m_tvnFolders.GetSelectedItem();			// Yes .. get handle to selected item.
		if (hItem != NULL)								// Was an item selected?
		{	
			tvItem.mask = TVIF_IMAGE | TVIF_STATE;		// Yes .. retrieve its image and state information.
			tvItem.hItem = hItem;
			m_tvnFolders.GetItem(&tvItem);
			if (tvItem.iImage == 1)						// Is item a bookmark?
			{
				*pResult = 0;							// Yes .. let control handle the keystroke.
			}
			else										// Not a bookmark, must be a folder.
			{
				nState = tvItem.state >> 12;			// Adjust value.
				if (nState > 1)							// If already checked,
					nState = 1;							// clear it.
				else									// Otherwise,
					nState = 2;							// check it.
				Set_State_Down(hItem, nState > 1);		// Set selection state for children.
				Set_State_Up(hItem);					// Adjust selection state for parents.
				*pResult = 1;							// Tell control we already handled the keystroke.
			}
		}
	}
	else					// Not the spacebar, so
		*pResult = 0;		// let the tree control handle it normally.
}


				///////////////////////////
				// Mouse event handlers. //
				///////////////////////////

// LMB released. Handles "drop" part of drag & drop.
void CSUView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	UINT nMyFlags;

	if (m_bIsDragging)							// Nothing to do unless user is dragging an item.
	{
		// Find out which item (if any) is under the mouse.
		m_hDragDest = m_tvnFolders.HitTest(point, &nMyFlags);
		m_pDragImage->DragLeave(this);			// Erase the drag image from the screen.
		m_pDragImage->EndDrag();				// End the drag operation.
		delete m_pDragImage;					// Delete the drag image.
		m_pDragImage = NULL;					// NULL its pointer.
		ReleaseCapture();						// Release mouse capture.
		m_bIsDragging = FALSE;					// Clear our flag.
		m_tvnFolders.SelectDropTarget(NULL);	// Clear tree control's drop target selection.
		if (m_hDragDest != NULL)				// Nothing to do if user didn't drop on an item.
		{
			if (m_hDragSource != m_hDragDest)	// Nothing to do if user dropped item on itself.
			{
				m_tvnFolders.SelectItem(m_hDragDest);	// Select the destination item.
				if (IsFolder(m_hDragSource))			// Call appropriate move routine based on item type.
					Move_Folder();
				else
					Move_Bookmark();
			}
		}
	}
	CFormView::OnLButtonUp(nFlags, point);
}

// RMB pressed. Used to cancel a drag operation.
void CSUView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	if (m_bIsDragging)						// If user is dragging an item,
	{
		m_bIsDragging = FALSE;				// Reset flag.
		m_pDragImage->DragLeave(this);		// Erase drag image from screen.
		m_pDragImage->EndDrag();			// End the drag.
		delete m_pDragImage;				// Delete the drag image.
		m_pDragImage = NULL;				// NULL its pointer.
		ReleaseCapture();					// Release mouse capture.
		m_tvnFolders.SelectDropTarget(NULL);	// Clear possible drop target selection.
		m_tvnFolders.EnsureVisible(m_hDragSource);	// Make sure item being dragged is visible.
	}
	CFormView::OnRButtonDown(nFlags, point);
}

// Mouse moved. Handles dragging.
void CSUView::OnMouseMove(UINT nFlags, CPoint point) 
{
	MSG			msg;
	UINT		nMyFlags,
				nMsg = 0;
	WPARAM		wParam;
	CPoint		ptAnchor,
				ptCurrent;
	HTREEITEM	hItem;

	if (m_bIsDragging)												// If user is dragging an item,
	{
		m_pDragImage->DragMove(point);								// Move the drag image.
		if ((hItem = m_tvnFolders.HitTest(point, &nMyFlags)) != NULL)	// Is mouse on a tree item?
		{
			m_pDragImage->DragLeave(this);					// If so, erase the drag image.
			m_tvnFolders.SelectDropTarget(hItem);			// Mark the target item.
			m_pDragImage->DragEnter(this, point);			// Redraw the drag image.
		}
		else	// Otherwise
		{
			// See if mouse is outside the tree control.
			if (nMyFlags & TVHT_ABOVE)		// Above the control?
			{
				nMsg = WM_VSCROLL;		// Vertical scrollbar
				wParam = SB_LINEUP;		// up one line.
			}
			else
			{
				if (nMyFlags & TVHT_BELOW)		// Below the control?
				{
					nMsg = WM_VSCROLL;		// Vertical scrollbar
					wParam = SB_LINEDOWN;	// down one line.
				}
				else
				{
					if (nMyFlags & TVHT_TOLEFT)	// Left of control?
					{
						nMsg = WM_HSCROLL;		// Horizontal scrollbar
						wParam = SB_LINELEFT;	// one line to the left.
					}
					else
					{
						if (nMyFlags & TVHT_TORIGHT)	// Right of control?
						{
							nMsg = WM_HSCROLL;		// Horizontal scrollbar
							wParam = SB_LINERIGHT;	// one line to the right.
						}
					}
				}
			}
			if (nMsg)		// Need to scroll?
			{
				ptAnchor = point;	// Save mouse position.
				do					// Repeat scrolling
				{
					m_pDragImage->DragLeave(this);							// Erase the drag image.
					::SendMessage(m_tvnFolders.m_hWnd, nMsg, wParam, NULL);	// Do it.
					m_pDragImage->DragEnter(this, point);					// Restore the drag image.
					GetCursorPos(&ptCurrent);		// Get mouse position.
					ScreenToClient(&ptCurrent);		// Convert to client coordinates
					// Check for pending mouse messages, but do not remove them from the message queue.
					if (PeekMessage(&msg, NULL, WM_MOUSEFIRST, WM_MOUSELAST, PM_NOREMOVE))
					{
						// If left button is released or right button is pressed,
						// stop scrolling and let the message be processed.
						if (msg.message == WM_LBUTTONUP || msg.message == WM_RBUTTONDOWN)
							break;
					}
					else
					{
						if (m_intScrollRate)			// If scroll rate is set,
							Sleep(m_intScrollRate);		// waste some time before next scroll event.
					}
				// If requested, continue scrolling until the mouse moves.
				} while (ptAnchor == ptCurrent);
			}
		}
	}
	CFormView::OnMouseMove(nFlags, point);
}

				/////////////////////////////////
				// CSUView "helper" functions. //
				/////////////////////////////////


// Fills the tree control from the current document.
// It is first called with the path_offset parm set to the root folder,
// then calls itself recursively to process all folders.
// Returns FALSE (and thus halts recursion) on any error.
BOOL CSUView::Process_Path(UINT nPathOffset, HTREEITEM hParentItem)
{
	UINT				i,
						nChild;
	CString				strTitle;
	CString				*pName;
	HTREEITEM			hLastItem;
	TV_INSERTSTRUCT		tvInsert;

	if (m_pMyDoc->m_nPathCount == 0)
		return FALSE;
	m_pMyDoc->m_pPathList[nPathOffset].bProcessed = TRUE;		// Mark current folder as processed.
	strTitle = *m_pMyDoc->m_pPathList[nPathOffset].pPath;		// Get path.
	m_pMyDoc->Friendly_Name(strTitle);							// Get name of last folder in path.
	tvInsert.item.pszText = strTitle.GetBuffer(strTitle.GetLength());
	tvInsert.item.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_STATE | TVIF_TEXT | TVIF_PARAM;
	tvInsert.item.state = INDEXTOSTATEIMAGEMASK(m_pMyDoc->m_pPathList[nPathOffset].nSelected);
	tvInsert.item.stateMask = TVIS_STATEIMAGEMASK;
	tvInsert.item.iSelectedImage = tvInsert.item.iImage = (m_pMyDoc->m_pPathList[nPathOffset].bLocked ? 2 : 0);
	tvInsert.item.lParam = nPathOffset;
	tvInsert.hParent = hParentItem;
	tvInsert.hInsertAfter = TVI_SORT;
	hLastItem = m_tvnFolders.InsertItem(&tvInsert);
	while(m_pMyDoc->Has_Children(nPathOffset, &nChild))	// Process any child folders.
	{
		if (!Process_Path(nChild, hLastItem))
			return FALSE;
	}
	// Now process any bookmarks in the current folder.
	for (i = 0; i < m_pMyDoc->m_nItemCount; i++)
	{
		if (m_pMyDoc->m_pDocElements[i].nPath == nPathOffset)
		{
			pName = m_pMyDoc->m_pDocElements[i].pPage;
			tvInsert.item.pszText = pName->GetBuffer(pName->GetLength());
			tvInsert.item.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_STATE | TVIF_TEXT | TVIF_PARAM;
			tvInsert.item.state = 0;
			tvInsert.item.stateMask = TVIS_STATEIMAGEMASK;
			tvInsert.item.iImage = tvInsert.item.iSelectedImage = 1;
			tvInsert.item.lParam = i;
			tvInsert.hParent = hLastItem;
			tvInsert.hInsertAfter = TVI_SORT;
			m_tvnFolders.InsertItem(&tvInsert);
		}
	}
	return TRUE;
}

// Returns TRUE if the specified tree item is a folder.
BOOL CSUView::IsFolder(HTREEITEM hItem)
{
	TV_ITEM tvItem;

	tvItem.mask = TVIF_IMAGE;
	tvItem.hItem = hItem;
	m_tvnFolders.GetItem(&tvItem);
	return (tvItem.iImage == 0 || tvItem.iImage == 2);
}

// Returns TRUE if the specified tree item is a bookmark.
BOOL CSUView::IsBookmark(HTREEITEM hItem)
{
	TV_ITEM tvItem;

	tvItem.mask = TVIF_IMAGE;
	tvItem.hItem = hItem;
	m_tvnFolders.GetItem(&tvItem);
	return (tvItem.iImage == 1);
}

// Called when: 
//    bRename = FALSE: User drops a bookmark on a valid location
//    bRename = TRUE:  User has renamed a bookmark.
// 
// Note that when bRename is TRUE, the OnEndlabeledit handler has
// already updated the document accordingly and set the m_hDragSource
// and m_hDragDest handles to the proper values. Thus, all we have to do
// is move the renamed item so it gets put into the tree in the proper sort order.
// This may seem complicated, but it's much faster than rebuilding the entire tree!
void CSUView::Move_Bookmark(BOOL bRename)
{
	char				buff1[MAX_PATH],
						buff2[MAX_PATH];
	UINT				nFolderOffset;
	TV_ITEM				tvItem;
	HTREEITEM			hNewItem = NULL,
						hTargetFolder = NULL,
						hNewSelection = NULL;
	TV_INSERTSTRUCT		tvInsert;

	if (m_hDragDest == NULL || m_hDragSource == NULL)
		return;

	// If we're not called as part of a rename operation,
	if (!bRename)
	{
		// Get handle of destination folder.
		if (IsFolder(m_hDragDest))
			hTargetFolder = m_hDragDest;
		else
			hTargetFolder = m_tvnFolders.GetParentItem(m_hDragDest);
		// If same as source folder, nothing to do.
		if (m_tvnFolders.GetParentItem(m_hDragSource) == hTargetFolder)
		{
			m_tvnFolders.SelectItem(m_hDragSource);
			return;
		}
	}
	else	// If we _are_ in a rename, handles have already been set.
	{
		hTargetFolder = m_hDragDest;
	}
	// Retrieve info for the bookmark being moved.
	// If "Follow moved bookmarks" option is not set,
	if (!m_bFollow)
	{
		// Determine which item to select after the move.
		hNewSelection = m_tvnFolders.GetNextItem(m_hDragSource, TVGN_NEXT);
		if (hNewSelection == NULL)
		{
			hNewSelection = m_tvnFolders.GetNextItem(m_hDragSource, TVGN_PREVIOUS);
			if (hNewSelection == NULL)
			{
				hNewSelection = m_tvnFolders.GetNextItem(m_hDragSource, TVGN_PREVIOUSVISIBLE);
				if (hNewSelection == NULL)
					hNewSelection = m_tvnFolders.GetNextItem(m_hDragSource, TVGN_NEXTVISIBLE);
			}
		}
	}
	tvInsert.item.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_CHILDREN | TVIF_STATE | TVIF_PARAM;
	tvItem.mask = TVIF_TEXT | TVIF_IMAGE;
	tvInsert.item.hItem = m_hDragSource;
	tvInsert.item.pszText = buff1;
	tvInsert.item.cchTextMax = sizeof(buff1);
	tvItem.pszText = buff2;
	tvItem.cchTextMax = sizeof(buff2);
	m_tvnFolders.GetItem(&tvInsert.item);
	tvInsert.item.iSelectedImage = tvInsert.item.iImage;
	// If we're not in a rename operation,
	if (!bRename)
	{
		// Scan parent folder for a bookmark with the same name.
		hNewItem = m_tvnFolders.GetChildItem(hTargetFolder);
		while(hNewItem != NULL)
		{
			tvItem.hItem = hNewItem;
			m_tvnFolders.GetItem(&tvItem);
			if (tvItem.iImage == 1)
			{
				// Give error message and bail if dupe found.
				if (stricmp(tvInsert.item.pszText, tvItem.pszText) == 0)
				{
					m_tvnFolders.EnsureVisible(hNewItem);
					AfxMessageBox(IDP_DUPE_BOOKMARK_NAME, MB_ICONINFORMATION);
					m_tvnFolders.SelectItem(m_hDragSource);
					return;
				}
			}
			hNewItem = m_tvnFolders.GetNextSiblingItem(hNewItem);
		}
		// Retrieve path offset for destination folder.
		nFolderOffset = m_tvnFolders.GetItemData(hTargetFolder);
		// Update document with source item's new parent folder.
		m_pMyDoc->m_pDocElements[tvInsert.item.lParam].nPath = nFolderOffset;
		// Mark document as dirty.
		m_pMyDoc->SetModifiedFlag();
	}
	// Insert the moved item into the tree.
	tvInsert.hParent = hTargetFolder;
	tvInsert.hInsertAfter = TVI_SORT;
	hNewItem = m_tvnFolders.InsertItem(&tvInsert);
	// Delete the original item.
	m_tvnFolders.DeleteItem(m_hDragSource);
	// If "Follow moved bookmarks" option is set,
	if (m_bFollow)
		hNewSelection = hNewItem;	// Selection follows the moved item.
	if (hNewSelection != NULL)
	{
		m_tvnFolders.EnsureVisible(hNewSelection);	// Make sure that appropriate item is visible
		m_tvnFolders.SelectItem(hNewSelection);		// and selected.
	}
	// Reset handles for drag/drop source and destination.
	m_hDragSource = NULL;
	m_hDragDest = NULL;
}

// This recursive function adjusts the bookmark offsets
// after a bookmark has been deleted from the document.
// It is first called with the handle of the root folder.
// It then calls itself with the handle of any subfolders
// found as it walks the current folder.
void CSUView::Adjust_Bookmarks(HTREEITEM parent)
{
	HTREEITEM	hCurItem;
	UINT		nOffset;

	hCurItem = m_tvnFolders.GetChildItem(parent);
	while (hCurItem != NULL)
	{
		if (IsFolder(hCurItem))
			Adjust_Bookmarks(hCurItem);
		else
		{
			nOffset = m_tvnFolders.GetItemData(hCurItem);
			if (nOffset >= m_nDeletedItem)
			{
				--nOffset;
				m_tvnFolders.SetItemData(hCurItem, nOffset);
			}
		}
		hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);
	}
}

// Tells whether or not the specified tree item can be moved.
BOOL CSUView::IsLocked(HTREEITEM item)
{
	TV_ITEM tvItem;

	tvItem.mask = TVIF_IMAGE;
	tvItem.hItem = item;
	m_tvnFolders.GetItem(&tvItem);
	return (tvItem.iImage >= 2);
}

// Handles folder move.
//    bRename = FALSE: User has dragged a folder to a new location.
//    bRename = TRUE: User has successfully renamed a folder.
//
// Note that when bRename = TRUE, the OnEndlabeledit handler has already
// taken care of updating the document and set the m_hDragSource and
// m_hDragDest member variables appropriately. Thus, this function simply
// moves the renamed folder so that is placed in the tree with the proper sort order.
void CSUView::Move_Folder(BOOL bRename)
{
	UINT		nSourceOffset,
				nDestOffset;
	TV_ITEM		tvItem;
	HTREEITEM	hTargetFolder,
				hScratch,
				hNewItem;

	if (m_hDragSource == NULL || m_hDragDest == NULL)
		return;
	// Unless move is part of a rename,
	if (!bRename)
	{
		// If drag destination was a bookmark,
		// get its parent folder.
		if (IsFolder(m_hDragDest))
			hTargetFolder = m_hDragDest;
		else
			hTargetFolder = m_tvnFolders.GetParentItem(m_hDragDest);
		// Nothing to do if folder was dropped on itself.
		if (m_hDragSource == hTargetFolder)
			return;
		// Nothing to do if folder was dropped on its own parent folder.
		hScratch = m_tvnFolders.GetParentItem(m_hDragSource);
		if (hScratch == hTargetFolder)
		{
			m_tvnFolders.SelectDropTarget(NULL);
			m_tvnFolders.SelectItem(m_hDragSource);
			return;
		}
		hScratch = m_tvnFolders.GetParentItem(hTargetFolder);
		// Make sure folder wasn't dropped on one of its own children.
		while (hScratch != NULL)
		{
			if (hScratch != m_hDragSource)
				hScratch = m_tvnFolders.GetParentItem(hScratch);
			else
			{
				AfxMessageBox(IDP_NO_CHILD_MOVE, MB_ICONINFORMATION);
				return;
			}
		}
		// Retrieve offsets for the source and destination folders.
		nSourceOffset = m_tvnFolders.GetItemData(m_hDragSource);
		nDestOffset = m_tvnFolders.GetItemData(hTargetFolder);
		// Tell document to adjust its path information, reporting error if it fails.
		if (!m_pMyDoc->Move_Folder(nSourceOffset, nDestOffset))
		{
			AfxMessageBox(IDP_DUPE_FOLDER_NAME, MB_ICONINFORMATION);
			return;
		}
		// Mark document as dirty.
		m_pMyDoc->SetModifiedFlag();
	}
	else
	{
		hTargetFolder = m_hDragDest;
	}
	// Get handle to source folder's parent. (Note: this handle will
	// never be NULL because the UI won't allow the root folder to be moved).
	hScratch = m_tvnFolders.GetParentItem(m_hDragSource);
	// Move the source folder to the destination folder in the tree,
	// and save handle to newly created folder.
	hNewItem = Move_Tree_Folder(m_hDragSource, hTargetFolder);
	m_tvnFolders.EnsureVisible(hNewItem);
	m_tvnFolders.SelectItem(hNewItem);
	// Delete the (now empty) source folder.
	if (!bRename)
	{
		// Get path offset for source folder's parent path.
		tvItem.mask = TVIF_PARAM;
		tvItem.hItem = hScratch;
		m_tvnFolders.GetItem(&tvItem);
		// Retrieve source folder's parent's selection state
		// (This was set appropriately by CSUDoc::Move_Folder)
		tvItem.state = (m_pMyDoc->m_pPathList[tvItem.lParam].nSelected) << 12;
		tvItem.stateMask = TVIS_STATEIMAGEMASK;
		tvItem.mask = TVIF_STATE;
		// Apply the proper selection state.
		m_tvnFolders.SetItem(&tvItem);
		// Adjust selection state for relevant parts of the tree.
		Set_State_Up(hNewItem);
		Set_State_Up(hScratch);
	}
	m_tvnFolders.DeleteItem(m_hDragSource);
	// Make sure the moved folder is visible and selected.
	m_hDragSource = NULL;
	m_hDragDest = NULL;
}

// Moves a folder item and all of its children
// to the specified destination folder.
// Returns handle to the newly created folder item.
HTREEITEM CSUView::Move_Tree_Folder(HTREEITEM hSource, HTREEITEM hDest)
{
	TCHAR				buff[_MAX_PATH];
	HTREEITEM			hNewItem,
						hFirstChild;
	TV_INSERTSTRUCT		tvi;

	// Retrieve information about source item.
	tvi.item.hItem = hSource;
	tvi.item.cchTextMax = sizeof(buff);
	tvi.item.pszText = buff;
	tvi.item.mask = TVIF_CHILDREN | TVIF_HANDLE | TVIF_IMAGE | TVIF_STATE | TVIF_TEXT | TVIF_PARAM;
	tvi.item.stateMask = TVIS_STATEIMAGEMASK;
	m_tvnFolders.GetItem(&tvi.item);
	// Insert a copy of the item into the target.
	// Note: Caller is responsible for deleting the original source item.
	tvi.hParent = hDest;
	// Place item in proper sorted position.
	tvi.hInsertAfter = TVI_SORT;
	tvi.item.mask = TVIF_IMAGE | TVIF_STATE | TVIF_TEXT | TVIF_PARAM;
	hNewItem = m_tvnFolders.InsertItem(&tvi);

	// Handle any child items.
	while ((hFirstChild = m_tvnFolders.GetChildItem(hSource)) != NULL)
	{
		Move_Tree_Folder(hFirstChild, hNewItem);    // recursively transfer all the items
		m_tvnFolders.DeleteItem(hFirstChild);		// delete the first child and all its children
	}
	// Return handle to the moved item.
	return hNewItem;
}

// Called recursively after user moves a folder.
// Ensures that the folder he just moved is displayed.
BOOL CSUView::Locate_Folder(HTREEITEM hParent, UINT nOffset)
{
	HTREEITEM hCurItem;

	if (nOffset == (m_tvnFolders.GetItemData(hParent)))	// If item offset matches path offset,
	{

		m_tvnFolders.EnsureVisible(hParent);	// Make sure tree is opened to the item.
		m_tvnFolders.SelectItem(hParent);		// Select the item.
		m_hCurItem = hParent;					// Save current item.
		Set_State_Up(hParent);					// Adjust selection state for parents.
		return TRUE;							// We're done.
	}
	if (!m_tvnFolders.ItemHasChildren(hParent))	// Keep looking if item has no children.
		return FALSE;
	hCurItem = m_tvnFolders.GetChildItem(hParent);	// Get first child item.
	while (hCurItem != NULL)						// Loop through remaining children.
	{
		if (IsFolder(hCurItem))						// If item is a folder,
			if (Locate_Folder(hCurItem, nOffset))	// Call ourselves.
				return TRUE;						// If that was the one, we're done.
		hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);	// Otherwise, move on to the next item.
	}
	return FALSE;
}

// Called recursively when a tree item's selection state changes.
// Sets or resets the selection state of specified item
// and all of its child folders.
void CSUView::Set_State_Down(HTREEITEM item, BOOL state)
{
	UINT		nItemState;
	TV_ITEM		tvItem;
	HTREEITEM	hCurItem;

	if (state)						// Convert from TRUE/FALSE to
		nItemState = 2 << 12;		// appropriate state mask value.
	else
		nItemState = 1 << 12;
	tvItem.mask = TVIF_STATE;
	tvItem.hItem = item;
	tvItem.state = nItemState;
	tvItem.stateMask = TVIS_STATEIMAGEMASK;
	m_tvnFolders.SetItem(&tvItem);			// Set current item state.
	tvItem.mask = TVIF_PARAM | TVIF_CHILDREN;	// Retrieve path offset and child flag.
	m_tvnFolders.GetItem(&tvItem);
	m_pMyDoc->m_pPathList[tvItem.lParam].nSelected = (state ? 2 : 1);	// Update selection state in the document.
	if (tvItem.cChildren)						// If item has children,
	{
		hCurItem = m_tvnFolders.GetChildItem(item);	// Get first child.
		while (hCurItem != NULL)
		{
			tvItem.mask = TVIF_IMAGE;			// Get item info.
			tvItem.hItem = hCurItem;
			m_tvnFolders.GetItem(&tvItem);
			if (tvItem.iImage != 1)				// If it's a folder,
				Set_State_Down(hCurItem, state);	// process it.
			hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);	// Next item.
		}
	}
}

// Called recursively when a tree item's selection state changes.
// Adjusts selection state of item's parents.
void CSUView::Set_State_Up(HTREEITEM hItem)
{
	char		szBuff[_MAX_PATH];
	BOOL		bAll,
				bNone;
	UINT		nItemState;
	TV_ITEM		tvItem;
	HTREEITEM	hCurItem,
				hParentItem;

	hParentItem = m_tvnFolders.GetParentItem(hItem);	// Get item's parent.
	if (hParentItem == NULL)							// Done if item is the root.
		return;
	bAll = TRUE;										// Set flags.
	bNone = TRUE;
	hCurItem = m_tvnFolders.GetChildItem(hParentItem);	// Walk through item's children.
	tvItem.stateMask = TVIS_STATEIMAGEMASK;
	tvItem.cchTextMax = MAX_PATH;
	tvItem.pszText = szBuff;
	while (hCurItem != NULL)
	{
		tvItem.mask = TVIF_IMAGE | TVIF_STATE | TVIF_TEXT;	// Retrieve item type & selection state.
		tvItem.hItem = hCurItem;
		m_tvnFolders.GetItem(&tvItem);
		if (tvItem.iImage != 1)					// Nothing to do if item is a bookmark.
		{
			if ((tvItem.state & TVIS_STATEIMAGEMASK) != 0x2000)	// If item is not fully selected,
				bAll = FALSE;									// clear flag and stop looking.
			if ((tvItem.state & TVIS_STATEIMAGEMASK)!= 0x1000)
				bNone = FALSE;
		}
		hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);	// Next item.
	}
	tvItem.mask = TVIF_PARAM | TVIF_STATE | TVIF_TEXT;			// Get item's path offset.
	tvItem.hItem = hParentItem;
	m_tvnFolders.GetItem(&tvItem);
	if (bNone)
	{
		if ((tvItem.state & TVIS_STATEIMAGEMASK) != 0x1000)
		{
			nItemState = 0x3000;
			m_pMyDoc->m_pPathList[tvItem.lParam].nSelected = 3;
		}
		else
			nItemState = tvItem.state;
	}
	else
	{
		if (bAll)	// If all children were selected,
					// then item is fully selected
			nItemState = (m_pMyDoc->m_pPathList[tvItem.lParam].nSelected = 2) << 12;
		else	// Otherwise,
				// item is partially selected.
			nItemState = (m_pMyDoc->m_pPathList[tvItem.lParam].nSelected = 3) << 12;
	}
	// Set selection state in the tree.
	tvItem.mask = TVIF_STATE;
	tvItem.state = nItemState;
	tvItem.stateMask = TVIS_STATEIMAGEMASK;
	m_tvnFolders.SetItem(&tvItem);
	// Proceed with next parent.
	Set_State_Up(hParentItem);
}

// Adjusts folder path offsets after a folder has been removed.
void CSUView::Adjust_Folder_Offsets(HTREEITEM item, UINT offset)
{
	TV_ITEM		tvItem;
	HTREEITEM	hChildItem,
				hCurItem;

	hCurItem = item;
	while (hCurItem != NULL)
	{
		tvItem.mask = TVIF_CHILDREN | TVIF_IMAGE | TVIF_PARAM;
		tvItem.hItem = hCurItem;
		m_tvnFolders.GetItem(&tvItem);
		if (tvItem.iImage != 1)
		{
			if (tvItem.cChildren != 0)
			{
				hChildItem = m_tvnFolders.GetChildItem(hCurItem);
				ASSERT(hChildItem != NULL);
				Adjust_Folder_Offsets(hChildItem, offset);
			}
			if ((UINT) (tvItem.lParam) >= offset)
			{
				--tvItem.lParam;
				tvItem.mask = TVIF_PARAM;
				m_tvnFolders.SetItem(&tvItem);
			}
		}
		hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);
	}
}

// Searches document for duplicate URLs and lets user choose which ones to keep.
BOOL CSUView::Weed_Duplicates()
{
	void				*pTemp;
	BOOL				bFound = FALSE,
						bBail = FALSE;
	UINT				i,
						j,
						nStart,
						nCount,
						nEnd,
						nMaxSets = 128,
						nSetCount = 0;
	CWeedDlg			dlg;
	CStatDlg			dlgStat;
	CString				strMessage;
	DUP_URL_SET			*pDupeSets = NULL;
	DUP_URL_ELEMENT		*pDupeItems = NULL;

	nCount = m_pMyDoc->m_nItemCount;				// Get number of bookmarks.
	pDupeItems = new DUP_URL_ELEMENT[nCount];		// Create array to hold info.
	pDupeSets = new DUP_URL_SET[nMaxSets];			// Create array to hold more infor.
	if (pDupeItems == NULL || pDupeSets == NULL)	// Bail if unable to create arrays.
	{
		AfxMessageBox(IDP_LOW_MEMORY, MB_ICONEXCLAMATION);
		return FALSE;
	}
	dlgStat.CenterWindow();		// Let the user know what's going on.
	dlgStat.Show_Message("Building working URL list...");
	// Load the info array from the document.
	for (i = 0; i < nCount; i++)
	{
		pDupeItems[i].pUrl = m_pMyDoc->m_pDocElements[i].pUrl;
		pDupeItems[i].nOffset = i;
		pDupeItems[i].bRemove = FALSE;
	}
	// Sort the array by URL.
	dlgStat.Show_Message("Sorting...");
	qsort((void*) pDupeItems, nCount, sizeof(DUP_URL_ELEMENT), Compare_Dupe_URLs);
	dlgStat.Show_Message("Searching...");
	// Walk the array and check for duplicates.
	for (i = 0; i < (nCount - 1); i++)
	{
		nStart = 0;		// Initialize offsets for current set of dupes.
		nEnd = 0;
		// Consecutive URLs match?
		if (pDupeItems[i].pUrl->Compare(*pDupeItems[i + 1].pUrl) == 0)
		{
			bFound = TRUE;		// Found a dupe.
			nStart = i;			// Save offsets for this set of dupes.
			nEnd = i + 1;
			while (nEnd < (nCount - 1))		// Loop through rest of array.
			{
				// Does next URL match?
				if (pDupeItems[nStart].pUrl->Compare(*pDupeItems[nEnd + 1].pUrl) == 0)
					++nEnd;	// Update end offset if so.
				else
					break;	// Done searching if not.
			}
			// Save data for current set of dupes.
			pDupeSets[nSetCount].nStart = nStart;
			pDupeSets[nSetCount].nEnd = nEnd;
			++nSetCount;
			// See if Dupe Set array needs to be bigger.
			if (nSetCount == nMaxSets)
			{
				// Try to make room for 32 more elements (this may leave some
				// "slop", but reduces the overhead that would occur if we reallocate
				// the array every time we expand to accomodate a single item).
				nMaxSets += 32;
				pTemp = realloc(pDupeSets, nMaxSets * sizeof(DUP_URL_SET));
				if (pTemp == NULL)
				{
					// If reallocation fails, report error and quit.
					AfxMessageBox(IDP_LOW_MEMORY, MB_ICONEXCLAMATION);
					bBail = TRUE;
					bFound = FALSE;
					break;
				}
				else
				{
					// Otherwise, set new array pointer.
					pDupeSets = (DUP_URL_SET*) pTemp;
				}
			}
		}
		if (nStart)		// If dupe was found,
			i = nEnd;	// Set loop counter past last dupe.
	}
	// Hide the status display for now.
	dlgStat.ShowWindow(SW_HIDE);
	// If dupes were found and no errors occurred,
	if (bFound && (!bBail))
	{
		dlg.m_pimgIcons = &m_imgTree;		// Give dialog pointers to our
		dlg.m_pimgBoxes = &m_imgState;		// image lists,
		dlg.m_pDupeItems = pDupeItems;		// duplicate array,
		dlg.m_pDupeSets = pDupeSets;		// duplicate set array,
		dlg.m_nSetCount = nSetCount;		// # of sets of dupes.
		dlg.m_pItemList = m_pMyDoc->m_pDocElements;	// Pointer to document data
		dlg.m_pPathList = m_pMyDoc->m_pPathList;	// and path data.
		bBail = (dlg.DoModal() != IDOK);		// Launch dialog, bBail if user cancels.
	}
	if (bBail)		// Did user bail?
	{
		bFound = FALSE;		// Yes .. Reset flag.
	}
	else
	{
		if (bFound)			// No .. Any dupes found?
		{
			if (dlg.m_nRemoveCount == 0)	// Yes .. Any selected for removal?
			{
				AfxMessageBox(IDP_NO_DUPES_REMOVED, MB_ICONINFORMATION);
				bFound = FALSE;
			}
			else
			{
				// Give user one last chance to bail.
				strMessage.Format("%u bookmarks will be removed.\n\nProceed?", dlg.m_nRemoveCount);
				if (AfxMessageBox(strMessage, MB_ICONQUESTION | MB_YESNO) != IDYES)
				{
					AfxMessageBox(IDP_COMMAND_CANCELLED, MB_ICONINFORMATION);
					bFound = FALSE;
				}
				else
				{
					// Let user know we're working on it.
					dlgStat.Show_Message("Removing marked items...");
					// Remove the marked items from the document.
					for (i = 0; i < nCount; i++)
					{
						if (pDupeItems[i].bRemove)
						{
							m_pMyDoc->Delete_Item(pDupeItems[i].nOffset);
							for (j = i + 1; j < nCount; j++)
							{
								if (pDupeItems[j].nOffset >= pDupeItems[i].nOffset)
									--pDupeItems[j].nOffset;
							}
						}
					}
				}
			}
		}
		else		// Let user know if no dupes exist.
			AfxMessageBox(IDP_NO_DUPES_FOUND, MB_ICONINFORMATION);
	}
	// Clean up.
	delete [] pDupeItems;
	delete [] pDupeSets;
	dlgStat.DestroyWindow();
	return bFound;
}

// Called recursively to expand or collapse all folders in the tree.
// Parms:
//    hItem = current item to process (NULL on first call, indicating root folder)
//    expand = flag; TRUE to expand, FALSE to collapse.
void CSUView::Expand_All(HTREEITEM hItem, BOOL bExpand)
{
	UINT		nCommand;
	BOOL		bIsRoot;
	TV_ITEM		tvItem;
	HTREEITEM	hCurItem,
				hThisItem;

	nCommand = bExpand ? TVE_EXPAND : TVE_COLLAPSE;	// Set desired command constant.
	if (hItem == NULL)								// Root folder?
	{
		hThisItem = m_tvnFolders.GetRootItem();		// Get its handle
		bIsRoot = TRUE;								// Set flag.
	}
	else
	{
		hThisItem = hItem;							// Use provided handle
		bIsRoot = FALSE;							// Clear flag.
	}
	tvItem.mask = TVIF_CHILDREN | TVIF_IMAGE;			// Retrieve image and child tvItem.
	tvItem.hItem = hThisItem;
	m_tvnFolders.GetItem(&tvItem);
	if (tvItem.cChildren)									// Item has children?
	{
		hCurItem = m_tvnFolders.GetChildItem(hThisItem);	// Get first child.
		do
		{
			tvItem.hItem = hCurItem;						// Get its info.
			m_tvnFolders.GetItem(&tvItem);
			if (tvItem.iImage != 1)							// Is it a folder?
				Expand_All(hCurItem, bExpand);				// Process it.
			hCurItem = m_tvnFolders.GetNextSiblingItem(hCurItem);	// Next item.
		}
		while (hCurItem != NULL);
	}
	if (bIsRoot)											// Root folder is always expanded.
	{
		m_tvnFolders.Expand(hThisItem, TVE_EXPAND);
		if (bExpand)
		{
			hCurItem = m_tvnFolders.GetSelectedItem();
			if (hCurItem != NULL)
				m_tvnFolders.EnsureVisible(hCurItem);
			else
			{
				m_tvnFolders.SelectItem(hThisItem);
				m_tvnFolders.EnsureVisible(hThisItem);
			}
		}
		else
			m_tvnFolders.SelectItem(hThisItem);
	}
	else
		m_tvnFolders.Expand(hThisItem, nCommand);
}

// Searches the tree, starting at the specified item,
// looking for a bookmark that matches the current search criteria.
// Called by OnEditFind (which sets the criteria) and
// OnEditFindNext (which continues the search with existing criteria).
// If matching bookmark is found, item is selected in the tree
// and function returns TRUE.
// Otherwise, existing selection remains unchanged and function returns FALSE.
BOOL CSUView::Search_Tree(HTREEITEM hStart)
{
	UINT		nOffset,
				nTestFunc;
	BOOL		bResult;
	TV_ITEM		tvItem;
	CString		strTarget(m_strSearchText);
	CString		strTest;
	HTREEITEM	hResult = NULL,
				hCurItem;

	strTarget.MakeLower();
	tvItem.mask = TVIF_IMAGE | TVIF_PARAM;		// Set mask to retrieve iImage and lParam
	// Determine which comparison to perform.
	nTestFunc = (m_intSearchIn * 2) + (m_bSearchCase ? 1 : 0);
	hCurItem = hStart;
	while (hCurItem != NULL)					// Search until no items left.
	{
		tvItem.hItem = hCurItem;				// Set handle to desired item.
		m_tvnFolders.GetItem(&tvItem);			// Retrieve item data.
		if (tvItem.iImage == 1)					// If item is a bookmark,
		{
			nOffset = (UINT) tvItem.lParam;		// Force item offset to a UINT.
			switch (nTestFunc)					// Compare appropriate fields.
			{
			case 0:								// Page and URL, case-sensitive.
				if (m_pMyDoc->m_pDocElements[nOffset].pPage->Find(m_strSearchText) != -1)
					bResult = TRUE;
				else
					bResult = m_pMyDoc->m_pDocElements[nOffset].pUrl->Find(m_strSearchText) != -1;
				break;
			case 1:								// Page and URL, case-insensitive.
				strTest = *m_pMyDoc->m_pDocElements[nOffset].pPage;
				strTest.MakeLower();
				if (strTest.Find(strTarget) != -1)
					bResult = TRUE;
				else
				{
					strTest = *m_pMyDoc->m_pDocElements[nOffset].pUrl;
					strTest.MakeLower();
					bResult = strTest.Find(strTarget) != -1;
				}
				break;
			case 2:								// Page only, case-sensitive.
				bResult = m_pMyDoc->m_pDocElements[nOffset].pPage->Find(m_strSearchText) != -1;
				break;
			case 3:								// Page only, case-insensitive.
				strTest = *m_pMyDoc->m_pDocElements[nOffset].pPage;
				strTest.MakeLower();
				bResult = strTest.Find(strTarget) != -1;
				break;
			case 4:								// URL only, case-sensitive.
				bResult = m_pMyDoc->m_pDocElements[nOffset].pUrl->Find(m_strSearchText) != -1;
				break;
			case 5:								// URL only, case-insensitive.
				strTest = *m_pMyDoc->m_pDocElements[nOffset].pUrl;
				strTest.MakeLower();
				bResult = strTest.Find(strTarget) != -1;
			}
			if (bResult)							// If item matched,
			{
				m_tvnFolders.EnsureVisible(hCurItem);	// Make sure it's visible.
				m_tvnFolders.SelectItem(hCurItem);		// Select it.
				return TRUE;							// Indicate success.
			}
		}
		hCurItem = Next_Tree_Item(hCurItem);			// Continue with next item.
	}
	return FALSE;		// Item not found, so return FALSE.
}

// Called by Search_Tree()
// Used to walk through the tree in top-down order,
// as if the entire tree were expanded.
// Input: Handle to current item.
// Returns: Handle to next item, NULL if there isn't one.
HTREEITEM CSUView::Next_Tree_Item(HTREEITEM hItem)
{
	HTREEITEM	hNextItem = NULL,
				hCurItem = hItem;

	hNextItem = m_tvnFolders.GetNextItem(hCurItem, TVGN_CHILD);		// Does item have any children?
	if (hNextItem != NULL)											// Return first child if so.
		return hNextItem;
	hNextItem = m_tvnFolders.GetNextItem(hCurItem, TVGN_NEXT);		// Any siblings left?
	if (hNextItem != NULL)											// Return sibling if so.
		return hNextItem;
	do
	{
		hCurItem = m_tvnFolders.GetNextItem(hCurItem, TVGN_PARENT);	// Get parent.
		if (hCurItem == NULL)										// Done if item has no parent.
			return NULL;
		else
			hNextItem = m_tvnFolders.GetNextItem(hCurItem, TVGN_NEXT);	// Get parent's next sibling.
	} while (hNextItem == NULL);	// Walk back up the tree until we either reach the root
	                                // or find a parent item that still has siblings.
	return hNextItem;
}

// Create new Netscape BOOKMARK.HTM file from the current document.
// Parm bBackup determines whether or not backup copy will be created.
// If not specified, defaults to TRUE.
void CSUView::Write_NS_File(BOOL bBackup)
{
	int				iPos;
	BOOL			bRenamed = FALSE;
	CString			strLine,
					strFileName,
					strBackupFileName;
	CStdioFile		file;
	CFileStatus		status;

	strFileName = m_pMyDoc->m_strNSPath;
	if (bBackup)								// If backup file requested,
	{
		iPos = strFileName.ReverseFind('.');		// Locate beginning of extension.
		strBackupFileName = strFileName.Left(iPos);	// Extract base name.
		strBackupFileName += ".bak";				// Append the ".bak" extension.
		status.m_ctime = 0;							// Initialize structure fields.
		status.m_size = 0;
		CFile::GetStatus(strBackupFileName, status);	// Attempt to retrieve file info.
		if (status.m_ctime == 0 && status.m_size == 0)	// If backup file doesn't exist,
		{
			if (Safe_Rename(strFileName, strBackupFileName, IDP_BAK_CREATE_FAIL))
				bRenamed = TRUE;
			else
				return;
		}
		else		// Otherwise, prompt user.
		{
			if (AfxMessageBox(IDP_QUERY_BAK_REPLACE, MB_YESNO | MB_ICONQUESTION) == IDYES)
			{
				try {									// Try to remove existing .bak file.
					CFile::Remove(strBackupFileName);
				}
				catch (CFileException* e) {				// Report error if Remove fails.
					e->Delete();
					AfxMessageBox(IDP_BAK_DELETE_FAIL, MB_ICONEXCLAMATION);
					return;
				}
				if (Safe_Rename(strFileName, strBackupFileName, IDP_BAK_CREATE_FAIL))
					bRenamed = TRUE;
				else
					return;
			}
		}
	}
	// Open file for writing, replacing existing contents.
	if (!file.Open(strFileName, CFile::modeCreate | CFile::modeWrite, NULL))
	{
		strLine.Format("Could not create or replace %s", strFileName);
		AfxMessageBox(strLine, MB_ICONINFORMATION);
		if (bBackup)
		{
			if (bRenamed)
				if (Safe_Rename(strBackupFileName, strFileName, IDP_BAK_RESTORE_FAIL))
					AfxMessageBox(IDP_BAK_RESTORE_OK, MB_ICONINFORMATION);
		}
		return;
	}
	m_pFile = &file;
	Write_NS_Folder(m_tvnFolders.GetRootItem());
	file.Close();
	m_pFile = NULL;
}

// Subroutine recurses through the tree
// and appends each folder's contents to BOOKMARK.HTM.
// Note that even though the file is opened in text mode,
// each line written is terminated with "\r\n", which actually puts
// a CR/CR/LF sequence at the end of each line. This seems bizarre, but
// that's actually the way Netscape writes the file!
void CSUView::Write_NS_Folder(HTREEITEM hThisFolder)
{
	char			*pBuff;
	UINT			nItemOffset;
	TV_ITEM			tvItem;
	HTREEITEM		hThisItem;
	static int		iIndent = 0;
	static char		szNameBuff[MAX_PATH];
	static CString	strLine;


	tvItem.mask = TVIF_IMAGE | TVIF_PARAM | TVIF_STATE | TVIF_TEXT;
	tvItem.stateMask = TVIS_STATEIMAGEMASK;
	tvItem.pszText = szNameBuff;
	tvItem.cchTextMax = sizeof(szNameBuff);

	// If this is the root directory,
	if (hThisFolder == m_tvnFolders.GetRootItem())
	{
		// Write the file header.
		strLine = "<!DOCTYPE NETSCAPE-Bookmark-file-1>\r\n"
				"<!-- This is an automatically generated file.\r\n"
				"It will be read and overwritten.\r\n"
				"Do Not Edit! -->\r\n";
		m_pFile->WriteString(strLine);
		strLine.Format("<TITLE>%s</TITLE>\r\n<H1>%s</H1>\r\n", m_pMyDoc->m_strNSPageTitle, m_pMyDoc->m_strNSPageTitle);
		m_pFile->WriteString(strLine);
	}
	tvItem.hItem = hThisFolder;
	m_tvnFolders.GetItem(&tvItem);
	// Done if folder is not selected.
//	if ((tvItem.state >> 12) < 2)
//		return;
	// Handle Netscape's indentions.
	pBuff = new char[iIndent + 5];
	memset(pBuff, 32, iIndent + 4);
	pBuff[iIndent] = '\0';
	// If folder is selected,
	if ((tvItem.state >> 12) >= 2)
	{
		// and it's not the root folder,
		if (hThisFolder != m_tvnFolders.GetRootItem())
		{
			// write the HTML tag for the folder name.
			strLine.Format("%s<DT><H3 FOLDED ADD_DATE=\"0\">%s</H3>\r\n", pBuff, szNameBuff);
			m_pFile->WriteString(strLine);
		}
	}
	// Write the HTML tag that marks the start of a folder's contents.
	// (This prevents an erroneous "fatal error message" when the
	// user exports with no folders selected at all.)
	strLine.Format("%s<DL><p>\r\n", pBuff);
	m_pFile->WriteString(strLine);
	// If folder is selected,
	if ((tvItem.state >> 12) >= 2)
	{
		// Adjust the indent.
		pBuff[iIndent] = ' ';
		iIndent += 4;
		pBuff[iIndent] = '\0';
		hThisItem = m_tvnFolders.GetChildItem(hThisFolder);
		while (hThisItem != NULL)
		{
			tvItem.hItem = hThisItem;
			m_tvnFolders.GetItem(&tvItem);
			if (tvItem.iImage != 1)
			{
				if ((tvItem.state >> 12) > 1)
					Write_NS_Folder(hThisItem);
			}
			else
			{
				nItemOffset = (UINT) tvItem.lParam;
				strLine.Format("%s<DT><A HREF=\"%s\" ADD_DATE=\"%u\" LAST_MODIFIED=\"%u\" LAST_VISIT=\"%u\">%s</A>\r\n",
							pBuff,
							*m_pMyDoc->m_pDocElements[nItemOffset].pUrl,
							m_pMyDoc->m_pDocElements[nItemOffset].nAdd,
							m_pMyDoc->m_pDocElements[nItemOffset].nMod,
							m_pMyDoc->m_pDocElements[nItemOffset].nVisit,
							szNameBuff);
				m_pFile->WriteString(strLine);
				// If item has a comment,
				if (m_pMyDoc->m_pDocElements[nItemOffset].pComment != NULL)
				{
					// Write the HTML for the comment to the file.
					// Note: Comment lines are taken from the file verbatim.
					// They are not indented, and the string itself contains any "extra"
					// CR's that were present in the original file.
					// For the record, Netscape does not even handle its own comments
					// gracefully when they contain any strLine breaks!
					strLine.Format("<DD>%s\n", *m_pMyDoc->m_pDocElements[nItemOffset].pComment);
					m_pFile->WriteString(strLine);
				}
			}
			hThisItem = m_tvnFolders.GetNextSiblingItem(hThisItem);
		}
		// Outdent after last folder item.
		iIndent -= 4;
		pBuff[iIndent] = '\0';
	}
	// Write HTML tag indicating end of folder.
	strLine.Format("%s</DL><p>\r\n", pBuff);
	m_pFile->WriteString(strLine);
	// Delete the buffer used for indents.
	delete [] pBuff;
}


// Formats a date value as a string.
// Input: Date value to be formatted
//        Reference to a CString object to contain the formatted date.
// If date is 0, string will contain "???".
// Otherwise, it will contain the "long" date formatted according to the
// system settings (via Control Panel | Regional Settings | Date tab). 
void CSUView::Format_Date_Field(time_t nDate, CString& strDate)
{
	if (nDate == 0)
		strDate = "???";
	else
	{
		CTime time(nDate);
		strDate = time.Format("%#c");
	}
}

// QSort callback function used by Weed_Duplicates()
int __cdecl Compare_Dupe_URLs(const void * p1, const void * p2)
{
	ASSERT(p1 != NULL && p2 != NULL);

	const DUP_URL_ELEMENT* i1 = (DUP_URL_ELEMENT*) p1;
	const DUP_URL_ELEMENT* i2 = (DUP_URL_ELEMENT*) p2;

	return (i1->pUrl->Compare(*i2->pUrl));
}


// File rename with exception handling.
// Used to create backup of BOOKMARK.HTM.
// Return value indicates success or failure.
BOOL CSUView::Safe_Rename(LPCTSTR pszOldName, LPCTSTR pszNewName, UINT nMsg)
{
	try
	{
		CFile::Rename(pszOldName, pszNewName);
	}
	catch (CFileException* e)
	{
		e->Delete();
		if (nMsg)
			AfxMessageBox(nMsg, MB_ICONEXCLAMATION);
		return FALSE;
	}
	return TRUE;
}


