// WeedDlg.cpp : implementation file
//
//
// SyncURLs Version 1.4
// Copyright  1998 Ziff-Davis Inc.
// First published in PC Magazine, US Edition
// January 6, 1998
// Written by Gregory A. Wolking
//

#include "stdafx.h"
#include "SyncURLs.h"
#include "MyList.h"
#include "WeedDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const char * szRootFolder = "Root Folder\\";
extern CSyncURLsApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CWeedDlg dialog


CWeedDlg::CWeedDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CWeedDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CWeedDlg)
	//}}AFX_DATA_INIT
}


void CWeedDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CWeedDlg)
	DDX_Control(pDX, IDC_lblHeader, m_lblHeader);
	DDX_Control(pDX, IDOK, m_cmdOK);
	DDX_Control(pDX, IDCANCEL, m_cmdCancel);
	DDX_Control(pDX, IDC_lstDupes, m_lstDupes);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CWeedDlg, CDialog)
	//{{AFX_MSG_MAP(CWeedDlg)
	ON_NOTIFY(LVN_GETDISPINFO, IDC_lstDupes, OnGetdispinfolstDupes)
	ON_WM_GETMINMAXINFO()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CWeedDlg message handlers

// This function modified for version 1.3.
// List Control text items now use the LPSTR_CALLBACK value, so the control
// obtains the text from the main application instead of storing its own copies
// of the individual strings. This reduces memory consumption during "Weed Duplicates"
// processing, and eliminates a crash problem when user has a very large number
// of duplicates.
BOOL CWeedDlg::OnInitDialog() 
{
	int			iItem = 0,
				nColumnWidth;
	UINT		i,
				j;
	BOOL		bZoomed;
	CRect		rectTarget, rectDesktop, rectMin(0, 0, 250, 150);
	CString		strLine;
	LV_ITEM		lvi;

	CDialog::OnInitDialog();
	// Retrieve saved window position.
	strLine = theApp.GetProfileString("Settings", "WeedDlgPos", "*");
	if (strLine != "*") // Skip if setting doesn't exist.
	{
		// Retrieve the numeric values.
		if (sscanf(strLine, "%i,%i,%i,%i,%i", &bZoomed, &rectTarget.left, &rectTarget.top, &rectTarget.right, &rectTarget.bottom) == 5)
		{
			MapDialogRect(&rectMin);
			// Adjust window position if needed, then move it.
			theApp.Adjust_Window_Pos(this, rectTarget, rectMin);
			if (bZoomed)					// if it was maximized,
				ShowWindow(SW_MAXIMIZE);	// maximize it now.
		}
	}
	m_lstDupes.SetImageList(m_pimgIcons, LVSIL_SMALL);				// Attach image lists to the
	m_lstDupes.SetImageList(m_pimgBoxes, LVSIL_STATE);				// List control.
	// Create list control's columns.
	m_lstDupes.InsertColumn(0, "URL / Folder Name", LVCFMT_LEFT, -1, -1);
	m_lstDupes.InsertColumn(1, "Page Name", LVCFMT_LEFT, -1, 1);
	// Now that list control is set up, fill it.
	// For each set of dupes:
	for (i = 0; i < m_nSetCount; i++)
	{
		// Set structure elements for URL item.
		lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_STATE | LVIF_PARAM;
		lvi.iItem = iItem;						// Item offset.
		lvi.iSubItem = 0;						// No sub-item.
		lvi.state = 0x5000;						// Display "URL" image instead of a checkbox.
		lvi.stateMask = LVIS_STATEIMAGEMASK;
		lvi.iImage = 3;							// Icon is a dash.
		lvi.pszText = LPSTR_TEXTCALLBACK;		// Query app for item text.
		lvi.cchTextMax = 0;
		// Set index to URL item, but force most significant bit to ON so that
		// callback function knows this item is a header.
		lvi.lParam = m_pDupeSets[i].nStart | 0x80000000;
		m_lstDupes.InsertItem(&lvi);			// Insert the item.
		++iItem;								// Next item index.
		lvi.state = 0x2000;						// Blank checkbox for bookmarks.
		lvi.iImage = 1;							// Icon for bookmarks.
		// For each item in the current set of dupes:
		for (j = m_pDupeSets[i].nStart; j <= m_pDupeSets[i].nEnd; j++)
		{
			// Set structure elements.
			lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_STATE | LVIF_PARAM;
			lvi.pszText = LPSTR_TEXTCALLBACK;	// Query app for item text.
			lvi.iSubItem = 0;			// Puts item in Column 0.
			lvi.iItem = iItem;
			lvi.lParam = j;				// Offset to item in dupe array.
			m_lstDupes.InsertItem(&lvi);
			// Set text only.
			lvi.mask = LVIF_TEXT;
			lvi.iSubItem = 1;			// Put text in column 1.
			lvi.pszText = LPSTR_TEXTCALLBACK;	// Query app for item text.
			m_lstDupes.SetItem(&lvi);	// Update the item.
			++iItem;					// Next item index.
		}
	}
	// Now that list has been filled, get size of client area.
	m_lstDupes.GetClientRect(&rectTarget);
	nColumnWidth = rectTarget.Width() / 2;
	// Set column widths accordingly.
	m_lstDupes.SetColumnWidth(0, nColumnWidth);
	m_lstDupes.SetColumnWidth(1, rectTarget.Width() - nColumnWidth);
	m_lstDupes.SetItemState(1, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
	return TRUE;
}

void CWeedDlg::OnOK() 
{
	LV_ITEM		lvi;
	UINT		nCount,
				i,
				j;
	BOOL		bKeepNone = TRUE;					// Flag for warning prompt.
	BOOL		bNotAll, bOneSet;

	m_nRemoveCount = 0;						// Initialize count.
	nCount = m_lstDupes.GetItemCount();		// Get # of items in the list.
	// Set structure elements to retrieve the data we need for each item.
	lvi.mask = LVIF_PARAM | LVIF_STATE | LVIF_IMAGE;
	lvi.iSubItem = 0;
	lvi.stateMask = LVIS_STATEIMAGEMASK;
	for (i = 0; i < nCount; i++)			// Walk through the list.
	{
		lvi.iItem = i;						// Get data for current item.
		m_lstDupes.GetItem(&lvi);
		if (lvi.iImage == 1)				// If item is a bookmark
		{
			if (lvi.state == 0x2000)		// and it is not selected,
			{
				++m_nRemoveCount;			// mark it for removal.
				m_pDupeItems[lvi.lParam].bRemove = TRUE;
			}
			else							// If item is selected,
			{                               // make sure it is _not_ marked for removal.
				m_pDupeItems[lvi.lParam].bRemove = FALSE;
				bKeepNone = FALSE;			// Clear "no items to keep" flag.
			}
		}
	}
	if (bKeepNone)	// If no items were selected, give user a chance to go back.
	{
		if (AfxMessageBox(IDP_REMOVE_ALL, MB_YESNO | MB_ICONQUESTION) != IDYES)
		{
			m_lstDupes.SetFocus();
			return;
		}
	}
	else // Otherwise, make sure at least one item in each set will be kept.
	{
		// Clear flag.
		bNotAll = FALSE;
		// For each set of dupes:
		for (i = 0; i < m_nSetCount && !bNotAll; i++)
		{
			// Set flag.
			bOneSet = TRUE;
			// For each item in the set,
			for (j = m_pDupeSets[i].nStart; (j <= m_pDupeSets[i].nEnd) && bOneSet; j++)
			{
				if (!m_pDupeItems[j].bRemove)	// If item is not marked for deletion,
					bOneSet = FALSE;			// Clear set flag.
			}
			// If we find a set that doesn't have at least one item
			// that is _not_ marked for removal, we're done searching.
			if (bOneSet)
				bNotAll = TRUE;
		}
		// If we found any sets with no items selected,
		// present an appropriate warning.
		if (bNotAll)
			if (AfxMessageBox(IDP_SET_UNMARKED,	MB_YESNO | MB_ICONQUESTION) != IDYES)
			{
				m_lstDupes.SetFocus();
				return;
			}
	}
	Save_Position();
	CDialog::OnOK();
}

void CWeedDlg::OnCancel() 
{
	m_nRemoveCount = 0;
	Save_Position();
	CDialog::OnCancel();
}

// This handler added for version 1.3 to support the use of the
// LPSTR_TEXTCALLBACK parameter for the list control's text strings. This
// lets the list control retrieve the text from our application's data storage
// instead of copying the strings and storing them in the control itself.

void CWeedDlg::OnGetdispinfolstDupes(NMHDR* pNMHDR, LRESULT* pResult) 
{
	UINT			nPath;
	const char		*pNewText = NULL;
	LV_DISPINFO		*pDispInfo = (LV_DISPINFO*)pNMHDR;
	const UINT		lParam = pDispInfo->item.lParam;

	// Is item a header?
	if (lParam & 0x80000000)
	{
		// Asking for column 0 text?
		if (pDispInfo->item.iSubItem == 0)
			// Point to URL text.
			pNewText = *m_pDupeItems[lParam & 0x7FFFFFFF].pUrl;
	}
	// If not a header, it must be a bookmark.
	else
	{
		switch (pDispInfo->item.iSubItem)		// Which column is control asking about?
		{
		case 0:			// Column 0 = item path string.
			nPath = m_pItemList[m_pDupeItems[lParam].nOffset].nPath;
			if (nPath == 0)		// Is path the Root Folder?
				pNewText = szRootFolder;		// Point to constant if so.
			else	// Otherwise, point to string in path array.
				pNewText = *m_pPathList[nPath].pPath;
			break;
		case 1:		// Column 1 = page title.
			pNewText = *m_pItemList[m_pDupeItems[lParam].nOffset].pPage;
			break;
		}
	}
	// Note: Under normal circumstances, it could be dangerous to cast pNewText as (char *)
	// instead of (const char *), as the pointer may point to the buffer belonging to 
	// one of our app's CString objects. However, we know that:
	// 1. The List Control's "Edit Labels" feature is disabled, so it will
	//    only be using this pointer to retrieve text for its screen display.
    // 2. The dialog itself does not modify any of those CString objects, so we don't have
	//    to worry about them moving to a different location in memory and rendering
	//    our pointer invalid.
	pDispInfo->item.pszText = (char *) pNewText;
	*pResult = 0;
}

// Called when dialog is resized.
void CWeedDlg::OnGetMinMaxInfo(MINMAXINFO FAR* lpMMI) 
{
	CRect rect(0, 0, 250, 150);					// Minimum size in DLUs.
	MapDialogRect(&rect);						// Convert DLUs to pixels.
	lpMMI->ptMinTrackSize.x = rect.right;		// Set minimum size.
	lpMMI->ptMinTrackSize.y = rect.bottom;
	CDialog::OnGetMinMaxInfo(lpMMI);
}

void CWeedDlg::OnSize(UINT nType, int cx, int cy) 
{
	CRect	rectMargins(0, 0, 4, 4),		// Size of dialog margins in DLUs.
			rectButtons(0, 0, 50, 14),		// Size of the OK and Cancel buttons in DLUs.
			rectClient,
			rectHeader,
			rectList,
			rectOK,
			rectCancel;

	if (m_lstDupes.m_hWnd != NULL)			// Don't try to adjust controls if they don't yet exist!
	{
		// Don't need to adjust controls unless we're being either maximized or restored.
		if (nType == SIZE_MAXIMIZED || nType == SIZE_RESTORED)
		{
			MapDialogRect(&rectMargins);				// Convert margins & button sizes from
			MapDialogRect(&rectButtons);				// DLUs to pixels.
			GetClientRect(&rectClient);					// Get size of new client rect.
			m_lblHeader.GetWindowRect(&rectHeader);		// Get current control positions.
			m_lstDupes.GetWindowRect(&rectList);
			m_cmdOK.GetWindowRect(&rectOK);
			m_cmdCancel.GetWindowRect(&rectCancel);
			ScreenToClient(&rectHeader);				// Convert to client coordinates.
			ScreenToClient(&rectList);
			ScreenToClient(&rectOK);
			ScreenToClient(&rectCancel);
			// Adjust right edge of the label and list control.
			rectHeader.right = rectClient.right - rectMargins.right;
			rectList.right = rectHeader.right;
			// Adjust bottom edge of list control.
			rectList.bottom = rectClient.bottom - rectButtons.bottom - rectMargins.bottom * 2;
			// Adjust top & bottom edges of OK and Cancel buttons.
			rectOK.top = rectCancel.top = rectClient.bottom - rectButtons.bottom - rectMargins.bottom;
			rectOK.bottom = rectCancel.bottom = rectClient.bottom - rectMargins.bottom;
			// Adjust left & right edges of the Cancel button.
			rectCancel.left = rectClient.right - rectButtons.right - rectMargins.right;
			rectCancel.right = rectClient.right - rectMargins.right;
			// Move controls to their new size and position.
			m_lblHeader.MoveWindow(&rectHeader, TRUE);
			// Don't know why, but the label doesn't redraw its text properly unless we force it to.
			m_lblHeader.RedrawWindow();
			m_lstDupes.MoveWindow(&rectList, TRUE);
			m_cmdOK.MoveWindow(&rectOK, TRUE);
			m_cmdCancel.MoveWindow(&rectCancel, TRUE);
		}
	}
	CDialog::OnSize(nType, cx, cy);
}

void CWeedDlg::Save_Position()
{
	BOOL	bZoomed;
	CRect	rect;
	CString strParms;
	
	if ((bZoomed = IsZoomed()))
		ShowWindow(SW_RESTORE);
	GetWindowRect(&rect);
	strParms.Format("%i,%i,%i,%i,%i", bZoomed, rect.left, rect.top, rect.right, rect.bottom);
	theApp.WriteProfileString("Settings", "WeedDlgPos", strParms);
}
